from openfisca_france_data.surveys import AbstractErfsSurveyScenario


class ErfsSurveyScenario(AbstractErfsSurveyScenario):
    used_as_input_variables = [
        'activite',
        'age_en_mois',
        'age',
        'autonomie_financiere',
        'categorie_salarie',
        'chomage_brut',
        'chomage_imposable',
        'contrat_de_travail',
        'cotisation_sociale_mode_recouvrement',
        'date_naissance',
        'effectif_entreprise',
        'f4ba',
        'heures_remunerees_volume',
        'hsup',
        'loyer',
        'nbF',
        'nbG',
        'nbH',
        'nbI',
        'nbJ',
        'nbN',
        'nbR',
        'pensions_alimentaires_percues',
        'rag',
        'retraite_brute',
        'retraite_imposable',
        'ric',
        'rnc',
        'salaire_de_base',
        'salaire_imposable',
        'statut_marital',
        'statut_occupation_logement',
        'taxe_habitation',
        'zone_apl',
        ]

    def __init__(self, year: int) -> None:
        self.year = year



from openfisca_france_data.surveys import AbstractErfsSurveyScenario


class ErfsFprSurveyScenario(AbstractErfsSurveyScenario):
    """Survey scenario spécialisé pour l'ERFS-FPR."""

    collection = "openfisca_erfs_fpr"

    # Les variables OpenFisca qu'on va utiliser avec les données en entrée.
    used_as_input_variables = [
        "activite",
        "autonomie_financiere",
        "categorie_salarie",
        "chomage_brut",
        "chomage_imposable",
        "contrat_de_travail",
        "cotisation_sociale_mode_recouvrement",
        "date_naissance",
        "effectif_entreprise",
        "f4ba",
        "heures_remunerees_volume",
        "loyer",
        "pensions_alimentaires_percues",
        "primes_fonction_publique",
        "rag",
        "retraite_brute",
        "retraite_imposable",
        "ric",
        "rnc",
        "statut_marital",
        "salaire_de_base",
        "statut_occupation_logement",
        "taxe_habitation",
        "traitement_indiciaire_brut",
        "zone_apl",
        ]

    # Might be used: hsup
    # Pour forcer quelques valeurs par défaut spécifiques.
    default_value_by_variable = dict(
        cotisation_sociale_mode_recouvrement = 2,
        # taux_incapacite = .50,
        )

    # En général, si les données contiennent les variables dont on a besoin, on les garde.
    #
    # Le cas échéant, on les neutralise à leur valeur par défaut, pour ne pas garder des
    # vecteurs des variables qu'on ne veut pas calculer.
    #
    # Ici c'est une exception : on déclare les variables qu'on ne veut pas neutraliser.
    non_neutralizable_variables = [
        "menage_ordinaire",
        "idfam_original",
        "idfoy_original",
        "idmen_original",
        # 'rempli_obligation_scolaire',
        # 'ressortissant_eee',
        "wprm_init",
        ]

    def __init__(self, year: int) -> None:
        self.year = year

    @classmethod
    def build_input_data(cls, year: int) -> None:
        # TODO: fix import, otherwise it is untestable.
        from openfisca_france_data.erfs_fpr.input_data_builder import build

        build(year = year)



import numpy as np

from openfisca_france.model.base import *  # noqa analysis:ignore

from openfisca_core.reforms import Reform
from openfisca_core.taxscales import MarginalRateTaxScale, combine_tax_scales

import logging

log = logging.getLogger(__name__)

TAUX_DE_PRIME = .10


class salaire_imposable_pour_inversion(Variable):
    value_type = float
    entity = Individu
    label = u'Salaire imposable utilisé pour remonter au salaire brut'
    definition_period = MONTH


class salaire_de_base(Variable):
    definition_period = MONTH

    def formula(individu, period, parameters):
        """Calcule le salaire brut à partir du salaire imposable par inversion du barème
        de cotisations sociales correspondant à la catégorie à laquelle appartient le salarié.
        """
        # Get value for year and divide below.
        this_year = period.this_year
        salaire_imposable_pour_inversion = individu('salaire_imposable_pour_inversion',
            period = this_year)

        # Calcule le salaire brut (salaire de base) à partir du salaire imposable.

        categorie_salarie = individu('categorie_salarie', period = this_year)
        contrat_de_travail = individu('contrat_de_travail', period = this_year)
        heures_remunerees_volume = individu('heures_remunerees_volume', period = this_year)
        hsup = individu('hsup', period = this_year)

        P = parameters(period)

        salarie = P.cotsoc.cotisations_salarie
        plafond_securite_sociale_annuel = P.prelevements_sociaux.pss.plafond_securite_sociale_mensuel * 12
        csg_deductible = parameters(period).prelevements_sociaux.contributions_sociales.csg.activite.deductible
        taux_csg = csg_deductible.taux
        taux_abattement = csg_deductible.abattement.rates[0]
        try:
            seuil_abattement = csg_deductible.abattement.thresholds[1]
        except IndexError:  # Pour gérer le fait que l'abattement n'a pas toujours était limité à 4 PSS
            seuil_abattement = None
        csg = MarginalRateTaxScale(name = 'csg')
        csg.add_bracket(0, taux_csg * (1 - taux_abattement))
        if seuil_abattement is not None:
            csg.add_bracket(seuil_abattement, taux_csg)

        target = dict()
        target['prive_non_cadre'] = set(['maladie', 'arrco', 'vieillesse_deplafonnee', 'vieillesse', 'agff', 'assedic'])
        target['prive_cadre'] = set(
            ['maladie', 'arrco', 'vieillesse_deplafonnee', 'agirc', 'cet', 'apec', 'vieillesse', 'agff', 'assedic']
            )

        for categorie in ['prive_non_cadre', 'prive_cadre']:
            baremes_collection = salarie[categorie]
            baremes_to_remove = list()
            for name, bareme in baremes_collection.items():
                if name.endswith('alsace_moselle'):
                    baremes_to_remove.append(name)
            for name in baremes_to_remove:
                del baremes_collection[name]

        for categorie in ['prive_non_cadre', 'prive_cadre']:
            test = set(
                name for name, bareme in salarie[categorie].items()
                if isinstance(bareme, MarginalRateTaxScale)
                )
            assert target[categorie] == test, 'target: {} \n test {}'.format(target[categorie], test)
        del bareme
        # On ajoute la CSG deductible et on proratise par le plafond de la sécurité sociale
        # Pour éviter les divisions 0 /0 dans le switch
        heures_remunerees_volume_avoid_warning = heures_remunerees_volume + (heures_remunerees_volume == 0) * 1e9
        salaire_imposable_proratise = switch(
            contrat_de_travail,
            {
                # temps plein
                0: salaire_imposable_pour_inversion / plafond_securite_sociale_annuel,
                # temps partiel
                1: salaire_imposable_pour_inversion / (
                    (heures_remunerees_volume_avoid_warning / (52 * 35)) * plafond_securite_sociale_annuel
                    ),
                }
            )
        salaire_de_base = 0.0
        for categorie in ['prive_non_cadre', 'prive_cadre']:
            bareme = combine_tax_scales(salarie[categorie])
            bareme.add_tax_scale(csg)
            brut_proratise = bareme.inverse().calc(salaire_imposable_proratise)
            assert np.isfinite(brut_proratise).all()
            brut = plafond_securite_sociale_annuel * switch(
                contrat_de_travail,
                {
                    # temps plein
                    0: brut_proratise,
                    # temps partiel
                    1: brut_proratise * (heures_remunerees_volume / (52 * 35)),
                    }
                )
            salaire_de_base += (
                (categorie_salarie == CATEGORIE_SALARIE[categorie]) * brut
                )
            assert (salaire_de_base > -1e9).all()
            assert (salaire_de_base < 1e9).all()
        # agirc_gmp
        # gmp = P.prelevements_sociaux.gmp
        # salaire_charniere = gmp.salaire_charniere_annuel
        # cotisation_forfaitaire = gmp.cotisation_forfaitaire_mensuelle_en_euros.part_salariale * 12
        # salaire_de_base += (
        #     (categorie_salarie == CATEGORIE_SALARIE['prive_cadre']) *
        #     (salaire_de_base <= salaire_charniere) *
        #     cotisation_forfaitaire
        #     )
        return salaire_de_base + hsup

        # public_titulaire_etat = salarie['public_titulaire_etat'] #.copy()
        # public_titulaire_etat['rafp'].multiply_rates(TAUX_DE_PRIME, inplace = True)
        # public_titulaire_etat = combine_tax_scales(salarie['public_titulaire_etat'])

# class traitement_indiciaire_brut(Variable):
#    definition_period = MONTH

#     def formula(self, simulation, period):
#         """Calcule le tratement indiciaire brut à partir du salaire imposable.
#         """
#         # Get value for year and divide below.
#         salaire_imposable_pour_inversion = simulation.calculate('salaire_imposable_pour_inversion',
#             period.start.offset('first-of', 'year').period('year'))

#         # Calcule le salaire brut à partir du salaire imposable.
#         # Sauf pour les fonctionnaires où il renvoie le traitement indiciaire brut
#         # Note : le supplément familial de traitement est imposable.
#         categorie_salarie = simulation.calculate('categorie_salarie', period)
#         P = simulation.parameters_at(period.start)
#         taux_csg = simulation.parameters_at(period.start).prelevements_sociaux.contributions_sociales.csg.activite.deductible.taux * (1 - .0175)
#         csg = MarginalRateTaxScale(name = 'csg')
#         csg.add_bracket(0, taux_csg)

#         salarie = P.cotsoc.cotisations_salarie

#         # public etat
#         # TODO: modifier la contribution exceptionelle de solidarité
#         # en fixant son seuil de non imposition dans le barème (à corriger dans param.xml
#         # et en tenant compte des éléments de l'assiette
#         # salarie['fonc']['etat']['excep_solidarite'] = salarie['fonc']['commun']['solidarite']

#         public_titulaire_etat = salarie['public_titulaire_etat']  # .copy()
#         public_titulaire_etat['rafp'].multiply_rates(TAUX_DE_PRIME, inplace = True)
#         public_titulaire_etat = combine_tax_scales(salarie['public_titulaire_etat'])

#         # public_titulaire_territoriale = combine_tax_scales(salarie['public_titulaire_territoriale'])
#         # public_titulaire_hospitaliere = combine_tax_scales(salarie['public_titulaire_hospitaliere'])
#         # public_non_titulaire = combine_tax_scales(salarie['public_non_titulaire'])

#         # Pour a fonction publique la csg est calculée sur l'ensemble TIB + primes
#         # Imposable = TIB - csg( (1+taux_prime)*TIB ) - pension(TIB) + taux_prime*TIB
#         bareme_csg_public_titulaire_etat = csg.multiply_rates(
#             1 + TAUX_DE_PRIME, inplace = False, new_name = "csg deduc titutaire etat")
#         public_titulaire_etat.add_tax_scale(bareme_csg_public_titulaire_etat)
#         bareme_prime = MarginalRateTaxScale(name = "taux de prime")
#         bareme_prime.add_bracket(0, -TAUX_DE_PRIME)  # barème équivalent à taux_prime*TIB
#         public_titulaire_etat.add_tax_scale(bareme_prime)
#         traitement_indiciaire_brut = (
#             (categorie_salarie == CATEGORIE_SALARIE['public_titulaire_etat']) *
#             public_titulaire_etat.inverse().calc(salaire_imposable_pour_inversion)
#             )
#         # TODO: complete this to deal with the fonctionnaire
#         # supp_familial_traitement = 0  # TODO: dépend de salbrut
#         # indemnite_residence = 0  # TODO: fix bug
#         return traitement_indiciaire_brut


# class primes_fonction_publique(Variable):
#    definition_period = MONTH

#     def formula(self, simulation, period):
#         """Calcule les primes.
#         """
#         # Get value for year and divide below.
#         traitement_indiciaire_brut = simulation.calculate('traitement_indiciaire_brut',
#             period.start.offset('first-of', 'year').period('year'))

#         return TAUX_DE_PRIME * traitement_indiciaire_brut


class inversion_directe_salaires(Reform):
    key = 'inversion_directe_salaires'
    name = u'Inversion des revenus'

    def apply(self):

        neutralized_reductions = [
            'accult',
            'adhcga',
            'assvie',
            'cappme',
            'cotsyn',
            'creaen',
            'daepad',
            'deffor',
            'dfppce',
            'doment',
            'domlog',
            'domsoc',
            'donapd',
            'duflot',
            'ecodev',
            'ecpess',
            'garext',
            'intagr',
            'intcon',
            'intemp',
            'invfor',
            'invlst',
            'invrev',
            'locmeu',
            'mecena',
            'mohist',
            'patnat',
            'prcomp',
            'repsoc',
            'resimm',
            'rsceha',
            'saldom',
            'scelli',
            'sofica',
            'sofipe',
            'spfcpi',
            ]
        neutralized_credits = [
            'accult',
            'acqgpl',
            'aidmob',
            'aidper',
            'assloy',
            'autent',
            'ci_garext',
            'cotsyn',
            'creimp',
            'creimp_exc_2008',
            'direpa',
            'divide',
            'drbail',
            'inthab',
            'jeunes',
            'mecena',
            'percvm',
            'preetu',
            'quaenv',
            'saldom2',
            ]

        neutralized_variables = [
            'avantage_en_nature',
            'complementaire_sante_employeur',
            'complementaire_sante_salarie',
            'exoneration_cotisations_employeur_apprenti',
            'exoneration_cotisations_employeur_stagiaire',
            'exoneration_cotisations_salariales_apprenti',
            'exoneration_cotisations_salarie_stagiaire',
            'indemnite_fin_contrat',
            'indemnites_compensatrices_conges_payes',
            'nouvelle_bonification_indiciaire',
            'primes_salaires',
            'prevoyance_obligatoire_cadre',
            'reintegration_titre_restaurant_employeur',
            'remuneration_apprenti',
            'rsa_non_calculable',
            'stage_gratification_reintegration',
            'residence_dom',
            'residence_guadeloupe',
            'residence_guyane',
            'residence_martinique',
            'residence_mayotte',
            'residence_reunion',
            # To reintegrate
            # Revenus
            'aah_base_ressources',
            'caah',
            'agirc_gmp_salarie',
            'supp_familial_traitement',  # Problème de l'autonomie financière
            'traitement_indiciaire_brut',
            ]
        neutralized_variables += list(set(neutralized_reductions + neutralized_credits))
        for neutralized_variable in neutralized_variables:
            log.info("Neutralizing {}".format(neutralized_variable))
            try:
                self.neutralize_variable(neutralized_variable)
            except AttributeError:
                continue

        self.add_variable(salaire_imposable_pour_inversion)
        for variable in [salaire_de_base]:  # traitement_indiciaire_brut, primes_fonction_publique,
            self.update_variable(variable)



#! /usr/bin/env python
import gc
import logging
import numpy as np

from openfisca_survey_manager.temporary import temporary_store_decorator  # type: ignore
from openfisca_france_data.erfs.input_data_builder.step_01_pre_processing import (
    create_variable_locataire,
    )
from openfisca_survey_manager.survey_collections import SurveyCollection  # type: ignore


log = logging.getLogger(__name__)


@temporary_store_decorator(file_name = "erfs_fpr")
def build_merged_dataframes(temporary_store = None, year = None):
    assert temporary_store is not None
    assert year is not None

    erfs_fpr_survey_collection = SurveyCollection.load(collection = "erfs_fpr")
    yr = str(year)[-2:]  # 12 for 2012
    yr1 = str(year+1)[-2:]  # 12 for 2012

    # where available, use harmoized data
    add_suffix_retropole_years = [2012]

    # infer names of the survey and data tables
    names = {
        "survey": f"erfs_fpr_{year}",
        "eec_individu": f"fpr_irf{yr}e{yr}t4" if year >= 2002 else f"fpr_irf{yr}e{yr1}",
        "eec_menage": f"fpr_mrf{yr}e{yr}t4" if year >= 2002 else f"fpr_mrf{yr}e{yr1}",
        "fpr_individu": f"fpr_indiv_{year}_retropole" if year in add_suffix_retropole_years else f"fpr_indiv_{year}",
        "fpr_menage": f"fpr_menage_{year}_retropole" if year in add_suffix_retropole_years else f"fpr_menage_{year}"
    }

    log.debug("Loading tables for year {} [{}]".format(year, names))

    # load survey and tables
    survey = erfs_fpr_survey_collection.get_survey(names['survey'])

    eec_individu = survey.get_values(table = names['eec_individu'], ignorecase= True)
    eec_menage = survey.get_values(table = names['eec_menage'], ignorecase=True)

    fpr_individu = survey.get_values(table = names['fpr_individu'], ignorecase = True)
    fpr_menage = survey.get_values(table = names['fpr_menage'], ignorecase = True)

    # transform to lowercase
    for table in (fpr_menage, eec_menage, eec_individu, fpr_individu):
        table.columns = [k.lower() for k in table.columns]

    # check column names prior to 2002
    if 'nopers' in eec_individu.columns:
        eec_individu.rename(columns = {'nopers':'noindiv'}, inplace = True)

    # merge EEC and FPR tables
    individus, menages = merge_tables(fpr_menage, eec_menage, eec_individu, fpr_individu, year)

    # check name of revenus fonciers variable
    if 'rev_fonciers' in menages.columns:
        log.info('Renaming rev_fonciers to rev_fonciers_bruts.')
        menages.rename(columns = {'rev_fonciers':'rev_fonciers_bruts'}, inplace = True)

    # store household table
    temporary_store[f"menages_{year}"] = menages
    del eec_menage, fpr_menage, menages
    gc.collect()

    # store individual-level table
    temporary_store[f"individus_{year}_post_01"] = individus
    del eec_individu, fpr_individu
    gc.collect()


def merge_tables(fpr_menage = None, eec_menage = None, eec_individu = None, fpr_individu = None, year = None,
        skip_menage = False):

    # Step 1: Individual-Level Data

    assert (eec_individu is not None) and (fpr_individu is not None)

    # Fusion enquête emploi et source fiscale
    nobs = {}
    nobs['fpr_ind'] = len(fpr_individu.noindiv.unique())
    nobs['eec_ind'] = len(eec_individu.noindiv.unique())

    # check name of 'acteu' variable and rename if necessary
    if 'act' in eec_individu.columns:
        eec_individu.rename(columns = {'act':'acteu'}, inplace = True)


    log.debug('There are {} obs. in the FPR individual-level data [unique(noindiv)]'.format(nobs['fpr_ind']))
    log.debug('There are {} obs. in the EEC individual-level data [unique(noindiv)]'.format(nobs['eec_ind']))

    log.debug('Columns in FPR-Ind table: {}.'.format(','.join(fpr_individu.columns)))
    log.debug('Columns in EEC-Ind table: {}.'.format(','.join(eec_individu.columns)))

    # merge tables
    individus = eec_individu.merge(fpr_individu, on = ['noindiv', 'ident', 'noi'], how = "inner")

    nobs['fpr_eec_ind'] = len(individus.noindiv.unique())
    log.debug('There are {} obs. in the FPR-EEC individual-level data [unique(noindiv)]'.format(nobs['fpr_eec_ind']))

    # check TBD
    check_naia_naim(individus, year)

    # establish list of variables, taking into account differences over time
    agepr = 'agepr' if year < 2013 else "ageprm"
    cohab = 'cohab' if year < 2013 else "coured"
    lien = 'lien' if year < 2013 else 'lienprm'  # TODO attention pas les mêmes modalités
    prosa = 'prosa' if year < 2013 else 'qprcent'  # TODO attention pas les mêmes modalités
    retrai = 'retrai' if year < 2013 else 'ret'  # TODO attention pas les mêmes modalités
    txtppb = 'txtpp' if year < 2004 else 'txtppb' if year < 2013 else 'txtppred'  # TODO attention pas les mêmes modalités
                                                                                # + pas utilisee (cf step_03 todo_create)
    acteu = 'act' if year < 2005 else 'acteu' # mêmes modalités (définition a changé)
    cstot = 'dcstot' if year < 2002 else 'cstotr' # mêmes modalité (0 = non-réponse)
    var_list = ([
        acteu,
        agepr,
        cohab,
        'contra',
        'forter',
        lien,
        'mrec',
        'naia',
        'noicon',
        'noimer',
        prosa,
        #retrai,
        'rstg',
        'statut',
        'stc',
        'titc',
        txtppb,
        ]
         + (["noiper"] if "noiper" in individus.columns else [])
         + (["encadr"] if "encadr" in individus.columns else [])
         + ([retrai] if retrai in individus.columns else []) #n'existe pas avant 2004
         + ([cstot] if cstot in individus.columns else [])) #existe 1996 - 2003, remplace retrai


    # fill NAs and type conversion
    for var in var_list:
        individus[var]=individus[var].fillna(0)
        individus[var]=individus[var].astype(np.int64)
        assert np.issubdtype(individus[var].dtype, np.integer), \
            "Variable {} dtype is {} and should be an integer".format(
                var, individus[var].dtype
                )

    # TBD
    if year >= 2013:
        individus['lpr'] = individus.lprm

    # Step 2: Household-Level Data
    if not skip_menage:

        assert (eec_menage is not None) and (fpr_menage is not None)

        nobs['fpr_men'] = len(fpr_menage.ident.unique())
        nobs['eec_men'] = len(eec_menage.ident.unique())

        log.debug('There are {} obs. in the FPR household-level data [unique(ident)]'.format(nobs['fpr_men']))
        log.debug('There are {} obs. in the EEC household-level data [unique(ident)]'.format(nobs['eec_men']))

        common_variables_pre = set(fpr_menage.columns).intersection(eec_menage.columns)

        if 'th' in common_variables_pre:
            fpr_menage.rename(columns = dict(th = 'taxe_habitation'), inplace = True)
            log.debug("Household-level tables: Renamed variable 'th' to 'taxe_habitation'")

        if 'tur5' in common_variables_pre:
            fpr_menage.drop('tur5', axis = 1, inplace = True)
            log.debug("Household-level tables: Variable 'tur5' has been removed from household-level data (FPR)")

        common_variables = set(fpr_menage.columns).intersection(eec_menage.columns)
        dropped_vars = common_variables_pre.symmetric_difference(common_variables)

        log.debug('Common variables in the household-level tables: [{}]'.format(','.join(common_variables)))

        if len(dropped_vars) > 0:
            log.debug('These household-level variables have been dropped: [{}]'.format(','.join(dropped_vars)))
        else:
            log.debug('No household-level variables have been dropped.')

        # merge FPR and EEC household-level data
        menages = fpr_menage.merge(eec_menage, on = 'ident', how = 'inner')

        nobs['fpr_eec_men'] = len(menages.ident.unique())
        log.debug('There are {} obs. in the FPR-EEC household-level data [unique(noindiv)]'.format(nobs['fpr_eec_men']))

        create_variable_locataire(menages)

        lprm = "lpr" if year < 2013 else "lprm"

        try:
            menages = menages.merge(
                individus.loc[individus[lprm] == 1, ["ident", "ddipl"]].copy()
                # lpr (ou lprm) == 1 ==> C'est la personne de référence
                )
        except Exception:
            print(individus.dtypes)
            raise

        nobs['merge_men'] = len(menages.ident.unique())
        nobs['merge_ind'] = len(menages.ident.unique())

        log.debug('There are {} individuals [before: {}] and {} households [before: {}] in the merged data table.'.format(nobs['merge_ind'], nobs['fpr_eec_ind'], nobs['merge_men'], nobs['fpr_eec_men']))


    # Infos sur les non appariés
    if not skip_menage:
        non_apparies(eec_individu, eec_menage, fpr_individu, fpr_menage)

    if skip_menage:
        menages = None

    m = individus.select_dtypes(np.number)
    individus[m.columns]= individus[m.columns].fillna(0)
    individus[m.columns]= individus[m.columns].astype('int64')

    m = menages.select_dtypes(np.number)
    menages[m.columns]= menages[m.columns].fillna(0)
    menages[m.columns]= menages[m.columns].astype('int64')

    return individus, menages


def non_apparies(eec_individu, eec_menage, fpr_individu, fpr_menage):
    """
    Ménages et individus non apparies
    """
    menages_non_apparies = eec_menage[
        ~(eec_menage.ident.isin(fpr_menage.ident.values))
        ].copy()
    individus_non_apparies = eec_individu[
        ~(eec_individu.ident.isin(fpr_individu.ident.values))
        ].copy()

    assert not menages_non_apparies.duplicated().any(), "{} menages sont dupliqués".format(
        menages_non_apparies.duplicated().sum())
    assert not individus_non_apparies.duplicated().any(), "{} individus sont dupliqués".format(
        individus_non_apparies.duplicated().sum())

    individus_non_apparies = individus_non_apparies.drop_duplicates(subset = 'ident', keep = 'last')
    difference = set(individus_non_apparies.ident).symmetric_difference(menages_non_apparies.ident)
    intersection = set(individus_non_apparies.ident) & set(menages_non_apparies.ident)

    log.debug("There are {} differences and {} intersections between the unmerged households and individuals.".format(len(difference), len(intersection)))

    del individus_non_apparies, menages_non_apparies, difference, intersection
    gc.collect()


def check_naia_naim(individus, year):
    valid_naim = individus.naim.isin(range(1, 13))
    if not valid_naim.all():
        log.debug('There are incorrect month or birth values (naim). They will be reset to 1.')

        individus.loc[
            individus.naim == 99,
            'naim'
            ] = 1  # np.random.randint(1, 13, sum(~valid_naim))
        individus.loc[
            individus.naim == 0,
            'naim'
            ] = 1
        individus.loc[
            individus.naim.isnull(),
            'naim'
            ] = 1

    assert individus.naim.isin(range(1, 13)).all(), f"naim values: {individus.naim.unique()}"
    assert isinstance(year, int)
    bad_noindiv = individus.loc[~((year >= individus.naia) & (individus.naia > 1890)),
                                "noindiv"].unique()
    for id in bad_noindiv:
        individus.loc[individus.noindiv == id,'naia'] = year - individus.loc[individus.noindiv == id, 'ageq']

    good = ((year >= individus.naia) & (individus.naia > 1890))
    assertion = good.all()
    bad_years = individus.loc[~good, "naia"].unique()
    bad_idents = individus.loc[~good, "ident"].unique()

    log.debug('There are incorrect years of birth [naia: {}] for individuals with ident [{}].'.format(';'.join([str(by) for by in bad_years]), ';'.join([str(bi) for bi in bad_idents])))

    try:
        lpr = "lpr" if year < 2013 else "lprm"
        lien = "lien" if year < 2013 else "lienprm"  # TODO attention pas les mêmes modalités
        prosa = "prosa" if year < 2013 else "qprcent"  # TODO attention pas les mêmes modalités
        retrai = "retrai" if year < 2013 else "ret"  # TODO attention pas les mêmes modalités
        assert assertion, "Error: \n {}".format(
            individus.loc[
                individus.ident.isin(bad_idents),  # WTF is this table supposed to be? I changed the 'lien' in lien
                # and so on for other variables
                [
                    'ag',
                    'ident',
                    lien,
                    'naia',
                    'naim',
                    'noi',
                    'noicon',
                    'noimer',
                    prosa,
                    retrai,
                    'rstg',
                    'statut',
                    'sexe',
                    lpr,
                    'chomage_i',
                    'pens_alim_recue_i',
                    'rag_i',
                    'retraites_i',
                    'ric_i',
                    'rnc_i',
                    'salaires_i',
                    ]
                    + (["noiper"] if "noiper" in individus.columns else [])
                ]
            )
    except AssertionError:
        if year == 2012:
            log.debug('Fixing erroneous years of birth [naia] manually for 2012.')
            individus.loc[
                (individus.ident == 12023304) & (individus.noi == 2),
                'naia'
                ] = 1954
            individus.loc[
                (individus.ident == 12041815) & (individus.noi == 1),
                'naia'
                ] = 2012 - 40
            #
        else:
            AssertionError('There have been issues with the year and month of birth (naia, naim) that need to be checked manually.')


if __name__ == '__main__':
    import sys
    import time
    start = time.time()
    logging.basicConfig(level = logging.DEBUG, stream = sys.stdout)
    # logging.basicConfig(level = logging.DEBUG, filename = 'run_all.log', filemode = 'w')
    year = 2014
    build_merged_dataframes(year = year)
    # TODO: create_enfants_a_naitre(year = year)
    log.info("Step 1 finished after {}".format(time.time() - start))
    print(time.time() - start)



#! /usr/bin/env python
import gc
import logging

from pandas import DataFrame, MultiIndex, concat
import numpy as np



from openfisca_france_data.erfs.input_data_builder.base import (
    year_specific_by_generic_data_frame_name)
from openfisca_survey_manager.temporary import temporary_store_decorator
from openfisca_survey_manager.survey_collections import SurveyCollection


log = logging.getLogger(__name__)


@temporary_store_decorator(file_name = 'erfs')
def create_fip(temporary_store = None, year = None):
    assert temporary_store is not None
    assert year is not None
    # FIP : fichier d'imposition des personnes
    """
    Creates a 'fipDat' table containing all these 'fip individuals'
    """
    # Some individuals are declared as 'personne à charge' (pac) on 'tax forms'
    # but are not present in the erf or eec tables.
    # We add them to ensure consistency between concepts.

    year_specific_by_generic = year_specific_by_generic_data_frame_name(year)

    erfs_survey_collection = SurveyCollection.load(
        collection = 'erfs', config_files_directory = config_files_directory)
    survey = erfs_survey_collection.get_survey('erfs_{}'.format(year))

    log.info("Démarrage de 03_fip")

    # anaisenf is a string containing letter code of pac (F,G,H,I,J,N,R) and year of birth (example: 'F1990H1992')
    # when a child is invalid, he appears twice in anaisenf (example: F1900G1900 is a single invalid child born in 1990)
    erfFoyVar = ['declar', 'anaisenf']
    foyer = survey.get_values(table = year_specific_by_generic["foyer"], variables = erfFoyVar)
    foyer.replace({'anaisenf': {'NA': np.nan}}, inplace = True)

    log.info("Etape 1 : on récupere les personnes à charge des foyers")
    log.info("    1.1 : Création des codes des enfants")
    foyer['anaisenf'] = foyer['anaisenf'].astype('string')
    nb_pac_max = len(max(foyer['anaisenf'], key=len)) / 5
    log.info("il ya a au maximum {} pac par foyer".format(nb_pac_max))

    # Separating the string coding the pac of each "déclaration".
    # Creating a list containing the new variables.

    # Creating the multi_index for the columns
    multi_index_columns = []
    assert int(nb_pac_max) == nb_pac_max, "nb_pac_max = {} which is not an integer".format(nb_pac_max)
    nb_pac_max = int(nb_pac_max)
    for i in range(1, nb_pac_max + 1):
        pac_tuples_list = [
            (i, 'declaration'),
            (i, 'type_pac'),
            (i, 'naia')
            ]
        multi_index_columns += pac_tuples_list

    columns = MultiIndex.from_tuples(
        multi_index_columns,
        names = ['pac_number', 'variable']
        )
    fip = DataFrame(np.random.randn(len(foyer), 3 * nb_pac_max), columns = columns)

    for i in range(1, nb_pac_max + 1):  # TODO: using values to deal with mismatching indexes
        fip[(i, 'declaration')] = foyer['declar'].values
        fip[(i, 'type_pac')] = foyer['anaisenf'].str[5 * (i - 1)].values
        fip[(i, 'naia')] = foyer['anaisenf'].str[5 * (i - 1) + 1: 5 * i].values

    fip = fip.stack("pac_number")
    fip.reset_index(inplace = True)
    fip.drop(['level_0'], axis = 1, inplace = True)

    log.info("    1.2 : elimination des foyers fiscaux sans pac")
    # Clearing missing values and changing data format
    fip = fip[(fip.type_pac.notnull()) & (fip.naia != 'an') & (fip.naia != '')].copy()
    fip = fip.sort(columns = ['declaration', 'naia', 'type_pac'])
    fip.set_index(["declaration", "pac_number"], inplace = True)
    fip = fip.reset_index()
    fip.drop(['pac_number'], axis = 1, inplace = True)
    assert fip.type_pac.isin(["F", "G", "H", "I", "J", "N", "R"]).all(), \
        "Certains types de PAC ne sont pas des cases connues"

    # control(fip, debug=True, verbose=True, verbose_columns=['naia'])

    log.info("    1.3 : on enlève les individus F pour lesquels il existe un individu G")
    type_FG = fip[fip.type_pac.isin(['F', 'G'])].copy()  # Filtre pour ne travailler que sur F & G

    type_FG['same_pair'] = type_FG.duplicated(subset = ['declaration', 'naia'], take_last = True)
    type_FG['is_twin'] = type_FG.duplicated(subset = ['declaration', 'naia', 'type_pac'])
    type_FG['to_keep'] = ~(type_FG['same_pair']) | type_FG['is_twin']
    # Note : On conserve ceux qui ont des couples déclar/naia différents et les jumeaux
    #       puis on retire les autres (à la fois F et G)
    fip['to_keep'] = np.nan
    fip.update(type_FG)
    log.info("    1.4 : on enlève les H pour lesquels il y a un I")
    type_HI = fip[fip.type_pac.isin(['H', 'I'])].copy()
    type_HI['same_pair'] = type_HI.duplicated(subset = ['declaration', 'naia'], take_last = True)
    type_HI['is_twin'] = type_HI.duplicated(subset = ['declaration', 'naia', 'type_pac'])
    type_HI['to_keep'] = (~(type_HI['same_pair']) | (type_HI['is_twin'])).values

    fip.update(type_HI)
    fip['to_keep'] = fip['to_keep'].fillna(True)
    log.info("{} F, G, H or I non redundant pac kept over {} potential candidates".format(
        fip['to_keep'].sum(), len(fip))
        )
    indivifip = fip[fip['to_keep']].copy()
    del indivifip['to_keep'], fip, type_FG, type_HI
    #
    # control(indivifip, debug=True)

    log.info("Step 2 : matching indivifip with eec file")
    indivi = temporary_store['indivim_{}'.format(year)]
    pac = indivi[(indivi.persfip.notnull()) & (indivi.persfip == 'pac')].copy()
    assert indivifip.naia.notnull().all(), "Il y a des valeurs manquantes de la variable naia"

    # For safety enforce pac.naia and indivifip.naia dtypes
    pac['naia'] = pac.naia.astype('int32')
    indivifip['naia'] = indivifip.naia.astype('int32')
    pac['key1'] = zip(pac.naia, pac['declar1'].str[:29])
    pac['key2'] = zip(pac.naia, pac['declar2'].str[:29])
    indivifip['key'] = zip(indivifip.naia.values, indivifip['declaration'].str[:29].values)
    assert pac.naia.dtype == indivifip.naia.dtype, \
        "Les dtypes de pac.naia {} et indvifip.naia {} sont différents".format(pac.naia.dtype, indivifip.naia.dtype)

    fip = indivifip[~(indivifip.key.isin(pac.key1.values))].copy()
    fip = fip[~(fip.key.isin(pac.key2.values))].copy()

    log.info("    2.1 new fip created")
    # We build a dataframe to link the pac to their type and noindiv
    tmp_pac1 = pac[['noindiv', 'key1']].copy()
    tmp_pac2 = pac[['noindiv', 'key2']].copy()
    tmp_indivifip = indivifip[['key', 'type_pac', 'naia']].copy()

    pac_ind1 = tmp_pac1.merge(tmp_indivifip, left_on='key1', right_on='key', how='inner')
    log.info("{} pac dans les 1ères déclarations".format(len(pac_ind1)))
    pac_ind2 = tmp_pac2.merge(tmp_indivifip, left_on='key2', right_on='key', how='inner')
    log.info("{} pac dans les 2èms déclarations".format(len(pac_ind2)))

    log.info("{} duplicated pac_ind1".format(pac_ind1.duplicated().sum()))
    log.info("{} duplicated pac_ind2".format(pac_ind2.duplicated().sum()))

    del pac_ind1['key1'], pac_ind2['key2']

    if len(pac_ind1.index) == 0:
        if len(pac_ind2.index) == 0:
            log.info("Warning : no link between pac and noindiv for both pacInd1&2")
        else:
            log.info("Warning : pacInd1 is an empty data frame")
            pacInd = pac_ind2
    elif len(pac_ind2.index) == 0:
        log.info("Warning : pacInd2 is an empty data frame")
        pacInd = pac_ind1
    else:
        pacInd = concat([pac_ind2, pac_ind1])
    assert len(pac_ind1) + len(pac_ind2) == len(pacInd)
    log.info("{} null pac_ind2.type_pac".format(pac_ind2.type_pac.isnull().sum()))
    log.info("pacInd.type_pac.value_counts()) \n {}".format(pacInd.type_pac.value_counts(dropna = False)))

    log.info("    2.2 : pacInd created")
    log.info("doublons noindiv, type_pac {}".format(pacInd.duplicated(['noindiv', 'type_pac']).sum()))
    log.info("doublons noindiv seulement {}".format(pacInd.duplicated('noindiv').sum()))
    log.info("nb de NaN {}".format(pacInd.type_pac.isnull().sum()))

    del pacInd["key"]
    pacIndiv = pacInd[~(pacInd.duplicated('noindiv'))].copy()
    # pacIndiv.reset_index(inplace=True)
    log.info("{}".format(pacIndiv.columns))

    temporary_store['pacIndiv_{}'.format(year)] = pacIndiv

    log.info("{}".format(pacIndiv.type_pac.value_counts()))
    gc.collect()

    # We keep the fip in the menage of their parents because it is used in to
    # build the famille. We should build an individual ident (ménage) for the fip that are
    # older than 18 since they are not in their parents' menage according to the eec
    log.info("{}".format(indivi['declar1'].str[0:2].value_counts()))
    log.info("{}".format(indivi['declar1'].str[0:2].describe()))
    log.info("{}".format(indivi['declar1'].str[0:2].notnull().all()))
    log.info("{}".format(indivi.info()))
    selection = indivi['declar1'].str[0:2] != ""
    indivi['noidec'] = indivi.declar1[selection].str[0:2].astype('int32')  # To be used later to set idfoy

    individec1 = indivi[(indivi.declar1.isin(fip.declaration.values)) & (indivi.persfip == "vous")]
    individec1 = individec1[["declar1", "noidec", "ident", "rga", "ztsai", "ztsao"]].copy()
    individec1 = individec1.rename(columns = {'declar1': 'declaration'})
    fip1 = fip.merge(individec1, on = 'declaration')
    log.info("    2.3 : fip1 created")

    individec2 = indivi.loc[
        (indivi.declar2.isin(fip.declaration.values)) & (indivi['persfip'] == "vous"),
        ["declar2", "noidec", "ident", "rga", "ztsai", "ztsao"]
        ].copy()
    individec2.rename(columns = {'declar2': 'declaration'}, inplace = True)
    fip2 = fip.merge(individec2)
    log.info("    2.4 : fip2 created")

    fip1.duplicated().value_counts()
    fip2.duplicated().value_counts()

    fip = concat([fip1, fip2])

    fip['persfip'] = 'pac'
    fip['year'] = year
    fip['year'] = fip['year'].astype('float')  # BUG; pas de colonne année dans la DF
    fip['noi'] = 99
    fip['noicon'] = None
    fip['noindiv'] = fip['declaration'].copy()
    fip['noiper'] = None
    fip['noimer'] = None
    fip['declar1'] = fip['declaration'].copy()
    fip['naim'] = 99
    fip['lien'] = None
    fip['quelfic'] = 'FIP'
    fip['acteu'] = None
    fip['agepf'] = fip['year'] - fip.naia.astype('float')
    fip['lpr'] = (fip['agepf'] <= 20) * 3 + (fip['agepf'] > 20) * 4
    fip['stc'] = None
    fip['contra'] = None
    fip['titc'] = None
    fip['mrec'] = None
    fip['forter'] = None
    fip['rstg'] = None
    fip['retrai'] = None
    fip['cohab'] = None
    fip['sexe'] = None
    fip['persfip'] = "pac"
    fip['agepr'] = None
    fip['actrec'] = (fip['agepf'] <= 15) * 9 + (fip['agepf'] > 15) * 5

    # TODO: probleme actrec des enfants fip entre 16 et 20 ans : on ne sait pas s'ils sont étudiants ou salariés */
    # TODO problème avec les mois des enfants FIP : voir si on ne peut pas remonter à ces valeurs: Alexis: clairement non

    # Reassigning noi for fip children if they are more than one per foyer fiscal
    fip["noi"] = fip["noi"].astype("int64")
    fip["ident"] = fip["ident"].astype("int64")

    fip_tmp = fip[['noi', 'ident']]

    while any(fip.duplicated(subset = ['noi', 'ident'])):
        fip_tmp = fip.loc[:, ['noi', 'ident']]
        dup = fip_tmp.duplicated()
        tmp = fip.loc[dup, 'noi']
        log.info("{}".format(len(tmp)))
        fip.loc[dup, 'noi'] = tmp.astype('int64') - 1

    fip['idfoy'] = 100 * fip['ident'] + fip['noidec']
    fip['noindiv'] = 100 * fip['ident'] + fip['noi']
    fip['type_pac'] = 0
    fip['key'] = 0

    log.info("Number of duplicated fip: {}".format(fip.duplicated('noindiv').value_counts()))
    temporary_store['fipDat_{}'.format(year)] = fip
    del fip, fip1, individec1, indivifip, indivi, pac
    log.info("fip sauvegardé")


if __name__ == '__main__':
    year = 2009
    logging.basicConfig(level = logging.INFO, filename = 'step_03.log', filemode = 'w')
    create_fip(year = year)
    log.info("etape 03 fichier des personnes imposables terminée")



#! /usr/bin/env python
import logging


import numpy
from pandas import merge, concat, DataFrame


from openfisca_france_data.erfs.input_data_builder.base import (
    year_specific_by_generic_data_frame_name)
from openfisca_france_data.utils import simulation_results_as_data_frame
from openfisca_france_data.erf import get_erf2of, get_of2erf
from openfisca_plugin_aggregates.aggregates import Aggregates
from openfisca_survey_manager.statshelpers import mark_weighted_percentiles as mwp
from openfisca_survey_manager.survey_collections import SurveyCollection


from openfisca_parsers import input_variables_extractors

log = logging.getLogger(__name__)


def clean(parameter):
    return parameter[:-len('_holder')] if parameter.endswith('_holder') else parameter


class Debugger(object):
    def __init__(self):
        super(Debugger, self).__init__()
        self.erf_menage = None
        self.erf_eec_indivi = None
        self.of_menages_data_frame = None
        self.of_individus_data_frame = None
        self.variable = None
        self.survey_scenario = None

    def set_survey_scenario(self, survey_scenario = None):
        assert survey_scenario is not None
        self.survey_scenario = survey_scenario
        self.variables = self.survey_scenario.simulation.tax_benefit_system.variables
        self.simulation = self.survey_scenario.simulation
        assert survey_scenario.simulation is not None, "The simulation attibute of survey_scenario is None"

    def set_variable(self, variable):
        if isinstance(variable, list):
            self.variable = variable[0]
        else:
            self.variable = variable

    def show_aggregates(self):
        from openfisca_france_data.erf.aggregates import build_erf_aggregates

        assert self.survey_scenario is not None, 'simulation attribute is None'
        assert self.variable is not None, 'variable attribute is None'
        variable = self.variable
        openfisca_aggregates = Aggregates()
        openfisca_aggregates.set_survey_scenario(self.survey_scenario)
        openfisca_aggregates.compute()

        variables = self.variables
        temp = (build_erf_aggregates(variables=[variable], year= self.survey_scenario.year))
        selection = openfisca_aggregates.aggr_frame["Mesure"] == variables[variable].label
        print(openfisca_aggregates.aggr_frame[selection])
        print(temp)
        # TODO: clean this
        return

    def extract(self, data_frame, entities = "men"):
        variables = self.variables
        filtered_data_frame_columns = list(set(variables.keys()).intersection(set(data_frame.columns)))
        extracted_columns = [column_name for column_name in filtered_data_frame_columns
                             if variables[column_name].entity in entities]
        extracted_columns = list(set(extracted_columns).union(set(['idmen'])))
        return data_frame[extracted_columns].copy()

    def get_all_parameters(self, column_list):
        global x
        print([column.name for column in column_list])
        x = x + 1
        if x == 20:
            boum
        variables = self.variables
        tax_benefit_system = self.survey_scenario.simulation.tax_benefit_system

        extractor = input_variables_extractors.setup(tax_benefit_system)

        if len(column_list) == 0:
            return []
        else:
            column_name = column_list[0].name
            print(column_name)
            if extractor.get_input_variables(variables[column_name]) is None:
                return column_list
            else:
                first_column = [column_list[0]]
                input_columns = self.get_all_parameters([
                    variables[clean(parameter)]
                    for parameter in list(extractor.get_input_variables(variables[column_name]))
                    ])
                other_columns = list(
                    set(self.get_all_parameters(column_list[1:])) - set(first_column + input_columns)
                    )
                print('input_variables: ', [column.name for column in input_columns])

                print('new_variables: ', [column.name for column in other_columns])

                new_column_list = first_column + input_columns + other_columns
                print('final list: ', [column.name for column in new_column_list])
                return new_column_list

    def build_columns_to_fetch(self):
        variables = self.variables
#        parameters_column = self.get_all_parameters([variables.get(x) for x in [self.variable]])
#        parameters = [x.name for x in parameters_column]
        parameters = [self.variable]
        # We want to get all parameters and consumers that we're going to encounter
#        consumers = []
#        for variable in [self.variable]:
#            column = variables.get(variable)
#            consumers = list(set(consumers).union(set(column.consumers)))
#        column_names = list(set(parameters).union(set(consumers)))

        # self.columns_to_fetch = column_names
        # self.variable_consumers = list(set(consumers))
        self.variable_parameters = list(set(parameters))
        self.columns_to_fetch = list(set(parameters))

    def build_openfisca_data_frames(self):
        column_names = self.columns_to_fetch
        for column in column_names:
            assert column in survey_scenario.tax_benefit_system.variables.keys()
        data_frame_by_entity_key_plural = survey_scenario.create_data_frame_by_entity(
            variables = column_names + ['idmen_original'],
            indices = True,
            roles = True,
            )
        self.data_frame_by_entity_key_plural = data_frame_by_entity_key_plural

        projected = self.project_on(data_frame_by_entity_key_plural = data_frame_by_entity_key_plural)
        idmen_original_by_idmen = dict(
            zip(
                data_frame_by_entity_key_plural['menages'].index.values,
                data_frame_by_entity_key_plural['menages']["idmen_original"].values
                )
            )
        self.idmen_original_by_idmen = idmen_original_by_idmen

        idmen_by_idmen_original = dict(
            zip(
                data_frame_by_entity_key_plural['menages']["idmen_original"].values,
                data_frame_by_entity_key_plural['menages'].index.values,
                )
            )
        self.idmen_by_idmen_original = idmen_by_idmen_original

        data_frame_by_entity_key_plural['menages'] = projected.rename(
            columns = {"idmen_original": "idmen"})
        data_frame_by_entity_key_plural['individus'].replace(
            {'idmen': idmen_original_by_idmen}, inplace = True)
        self.data_frame_by_entity_key_plural = data_frame_by_entity_key_plural

    def project_on(self, receiving_entity_key_plural = 'menages', data_frame_by_entity_key_plural = None):
        tax_benefit_system = self.survey_scenario.tax_benefit_system
        assert data_frame_by_entity_key_plural is not None
        assert receiving_entity_key_plural is not tax_benefit_system.person_key_plural

        entity_data_frame = data_frame_by_entity_key_plural[receiving_entity_key_plural]
        person_data_frame = data_frame_by_entity_key_plural[tax_benefit_system.person_key_plural]

        entity_keys_plural = list(
            set(tax_benefit_system.entity_class_by_key_plural.keys()).difference(set(
                [tax_benefit_system.person_key_plural, receiving_entity_key_plural]
                ))
            )

        for entity_key_plural in entity_keys_plural:
            entity = tax_benefit_system.entity_class_by_key_plural[entity_key_plural]
            # Getting only heads of other entities prenent in the projected on entity
            boolean_index = person_data_frame[entity.role_for_person_variable_name] == 0  # Heads
            index_entity = person_data_frame.loc[boolean_index, entity.index_for_person_variable_name].values  # Ent.
            for column_name, column_series in self.data_frame_by_entity_key_plural[entity_key_plural].items():
                person_data_frame.loc[boolean_index, column_name] \
                    = column_series.iloc[index_entity].values
                person_data_frame[column_name].fillna(0)

        receiving_entity = tax_benefit_system.entity_class_by_key_plural[receiving_entity_key_plural]
        grouped_data_frame = person_data_frame.groupby(by = receiving_entity.index_for_person_variable_name).agg(sum)
        grouped_data_frame.drop(receiving_entity.role_for_person_variable_name, axis = 1, inplace = True)
        data_frame = concat([entity_data_frame, grouped_data_frame], axis = 1)

        assert data_frame.notnull().all().all()
        return data_frame

    def build_erf_data_frames(self):
        # TODO: remove this
        self.columns_to_fetch = ['af']
        variables = self.columns_to_fetch
        erf_survey_collection = SurveyCollection.load(
            collection = "erfs", config_files_directory = config_files_directory)
        erf_survey = erf_survey_collection.get_survey("erfs_{}".format(year))
        year_specific_by_generic = year_specific_by_generic_data_frame_name(year)
        generic_by_year_specific = dict(zip(year_specific_by_generic.values(), year_specific_by_generic.keys()))

        erf_variables = list(set(variables + ["ident", "wprm", "quelfic", "noi"]))
        of2erf = get_of2erf()
        for index, variable in enumerate(erf_variables):
            if variable in of2erf:
                erf_variables[index] = of2erf[variable]
        data_frame_by_table = dict(eec_indivi = None, erf_indivi = None, erf_menage = None)
        erf_variables_by_generic_table = dict(eec_indivi = [], erf_indivi = [], erf_menage = [])

        year_specific_tables_by_erf_variable = dict(
            [
                (
                    erf_variable,
                    set(
                        erf_survey.find_tables(variable = erf_variable)
                        ).intersection(
                        set([year_specific_by_generic[key] for key in erf_variables_by_generic_table.keys()])
                        )
                    ) for erf_variable in erf_variables
                ]
            )
        for variable, year_specific_tables in year_specific_tables_by_erf_variable.items():
            if len(year_specific_tables) < 1:
                log.info("No tables are present for variable {}".format(variable))
                continue
            else:
                log.info("Variable {} is present in multiple tables : {}".format(variable, year_specific_tables))
                for table in year_specific_tables:
                    log.info("Variable {} is retrieved from table {}".format(variable, table))
                    erf_variables_by_generic_table[generic_by_year_specific[table]].append(variable)

        erf2of = get_erf2of()

        for table, erf_variables in erf_variables_by_generic_table.items():
            if erf_variables:
                data_frame_by_table[table] = erf_survey.get_values(
                    variables = erf_variables, table = year_specific_by_generic[table]
                    )
                data_frame_by_table[table].rename(columns = erf2of, inplace = True)
                data_frame_by_table[table].rename(columns = {'ident': 'idmen'}, inplace = True)

        assert not data_frame_by_table["erf_menage"].duplicated().any(), "Duplicated idmen in erf_menage"
        self.erf_data_frame_by_entity_key_plural = dict(
            menages = data_frame_by_table["erf_menage"],
            individus = data_frame_by_table["erf_indivi"].merge(data_frame_by_table["eec_indivi"])
            )
    # TODO: fichier foyer

    def get_major_differences(self):
        variable = self.variable

        of_menages_data_frame = self.data_frame_by_entity_key_plural['menages']
        erf_menages_data_frame = self.erf_data_frame_by_entity_key_plural['menages']

        merged_menage_data_frame = merge(
            erf_menages_data_frame[[variable, 'idmen']],
            of_menages_data_frame[[variable, 'idmen']],
            on = 'idmen',
            how = 'inner',
            suffixes = ('_erf', '_of')
            )

        log.info('Length of merged_menage_data_frame is {}'.format(len(merged_menage_data_frame)))
        merged_menage_data_frame.set_index('idmen', drop = False, inplace = True)
        table = merged_menage_data_frame[
            numpy.logical_and(
                merged_menage_data_frame[variable + '_erf'] != 0,
                merged_menage_data_frame[variable + '_of'] != 0
                )
            ]
        table[variable + "_rel_diff"] = (table[variable + '_of'] - table[variable + '_erf']) \
            / table[variable + '_erf']  # Difference relative
        log.info(
            "Minimum difference between the two tables for {} is {}".format(
                variable, str(table[variable + "_rel_diff"].min())
                )
            )
        log.info(
            "Maximum difference between the two tables for {} is {}".format(
                variable, str(table[variable + "_rel_diff"].max())
                )
            )
        table[variable + '_ratio'] = (
            table[variable + '_of'] / table[variable + '_erf']
            )
        log.info(table[variable + "_rel_diff"].describe())
        try:
            assert len(table[variable + "_rel_diff"]) == len(table['wprm_of']), "PINAGS"
            dec, values = mwp(
                table[variable + "_rel_diff"],
                numpy.arange(1, 11), table['wprm_of'],
                2,
                return_quantiles = True
                )
            log.info(sorted(values))
            dec, values = mwp(
                table[variable + "_rel_diff"],
                numpy.arange(1, 101),
                table['wprm_erf'],
                2,
                return_quantiles = True
                )
            log.info(sorted(values)[90:])
            del dec, values
        except Exception:
            log.info('Weighted percentile method did not work for {}'.format(variable + "_rel_diff"))
            pass
        table.sort(columns = variable + "_rel_diff", ascending = False, inplace = True)

        print(table[:10].to_string())
        return table

    def describe_discrepancies(self, fov = 10, consumers = False, parameters = True, descending = True, to_men = False):
        variable = self.variable
        major_differences_data_frame = self.get_major_differences()
        major_differences_data_frame.sort(
            columns = self.variable + "_rel_diff",
            ascending = not descending,
            inplace = True
            )
        debug_data_frame = major_differences_data_frame[0:fov].copy()
        del major_differences_data_frame

        of_menages_data_frame = self.data_frame_by_entity_key_plural['menages']
        of_individus_data_frame = self.data_frame_by_entity_key_plural['individus']
        erf_individus_data_frame = self.erf_data_frame_by_entity_key_plural['individus']
        erf_menages_data_frame = self.erf_data_frame_by_entity_key_plural['menages']
        return debug_data_frame

        kept_columns = set()
        if parameters:
            kept_columns.update(set(self.variable_parameters))
        if consumers:
            kept_columns.update(set(self.variable_consumers))
        kept_columns = list(kept_columns)
        kept_columns = list(set(kept_columns).union(
            set(['idmen', 'idfam', 'idfoy', 'quimen', 'quifam', 'quifoy'] + list(major_differences_data_frame.columns)))
            )

        if to_men:
            entities_ind = ['ind']
            entities_men = ['men', 'fam', 'foy']
        else:
            entities_ind = ['ind', 'fam', 'foy']
            entities_men = ['men']

        debug_data_frame = debug_data_frame.merge(
            self.extract(of_menages_data_frame, entities = entities_men),
            how = 'inner',
            on = 'idmen',
            )

        print(debug_data_frame.to_string())

        debug_data_frame = debug_data_frame.merge(
            self.extract(of_individus_data_frame, entities = entities_ind),
            how = 'inner',
            on = 'idmen',
            )

        debug_data_frame = debug_data_frame.merge(
            erf_individus_data_frame,
            how = 'inner',
            on = 'idmen',
            )

        suffixes = ["_erf", "_of", "_rel_diff", "_ratio"]
        reordered_columns = [variable + suffixe for suffixe in suffixes] \
            + ["idmen", "quimen", "idfam", "quifam", "idfoy", "quifoy"]
        reordered_columns = reordered_columns + list(set(kept_columns) - set(reordered_columns))
        debug_data_frame = debug_data_frame[reordered_columns].copy()
        return debug_data_frame

    def generate_test_case(self):
        entity_class_by_key_plural = self.survey_scenario.tax_benefit_system.entity_class_by_key_plural
        menages_entity = entity_class_by_key_plural['menages']
        idmen_by_idmen_original = self.idmen_by_idmen_original
        idmen_original = self.describe_discrepancies(descending = False)[
            menages_entity.index_for_person_variable_name].iloc[0]
        idmen = idmen_by_idmen_original[idmen_original]
        input_data_frame = self.survey_scenario.input_data_frame
        individus_index = input_data_frame.index[
            input_data_frame[menages_entity.index_for_person_variable_name] == idmen]
        index_by_entity = {
            entity_class_by_key_plural['individus']: individus_index,
            }
        for entity in list(entity_class_by_key_plural.values()):
            if entity.key_plural != 'individus':
                index_by_entity[entity] = input_data_frame.loc[
                    individus_index, entity.index_for_person_variable_name].unique()

        extracted_indices = individus_index
        for entity, entity_index in index_by_entity.items():
            if entity.key_plural in ['menages', 'individus']:
                continue
            extracted_indices = extracted_indices + \
                input_data_frame.index[input_data_frame[entity.index_for_person_variable_name].isin(entity_index)]

        extracted_input_data_frame = input_data_frame.loc[extracted_indices]
        return extracted_input_data_frame


if __name__ == '__main__':
    import sys
    from openfisca_plugin_aggregates.tests.test_aggregates import create_survey_scenario
    logging.basicConfig(level = logging.INFO, stream = sys.stdout)
    restart = True
    if restart:
        year = 2009
        survey_scenario = create_survey_scenario(year)
        survey_scenario.simulation = survey_scenario.new_simulation()

    debugger = Debugger()
    debugger.set_survey_scenario(survey_scenario = survey_scenario)
    debugger.set_variable('af')
    debugger.build_columns_to_fetch()
    debugger.build_openfisca_data_frames()
    debugger.build_erf_data_frames()

    # df_menage = debugger.data_frame_by_entity_key_plural['menages']
    # df_famille = debugger.data_frame_by_entity_key_plural['familles']
    # df_individus = debugger.data_frame_by_entity_key_plural['individus']

    # df = debugger.get_major_differences()

    # debugger.show_aggregates()
    df = debugger.describe_discrepancies(descending = False)
    df = debugger.generate_test_case()

    boum
    entity_class_by_key_plural = debugger.survey_scenario.tax_benefit_system.entity_class_by_key_plural
    menages_entity = entity_class_by_key_plural['menages']

    idmen = debugger.describe_discrepancies(descending = False)[menages_entity.index_for_person_variable_name].iloc[0]

    input_data_frame = debugger.survey_scenario.input_data_frame



from openfisca_france_data.utils import build_cerfa_fields_by_column_name  # type: ignore


def test_build_cerfa_fields_by_column_name():
    cerfa_fields_by_column_name_2006 = build_cerfa_fields_by_column_name(2006, [6, 7, 8])
    cerfa_fields_by_column_name_2009 = build_cerfa_fields_by_column_name(2006, [6, 7, 8])
    cerfa_fields_by_column_name_2012 = build_cerfa_fields_by_column_name(2012, [6, 7, 8])

    assert cerfa_fields_by_column_name_2006['f6ss'] == ['f6ss', 'f6st', 'f6su']
    assert cerfa_fields_by_column_name_2009['f6ss'] == ['f6ss', 'f6st', 'f6su']
    assert cerfa_fields_by_column_name_2012['f6ss'] == ['f6ss', 'f6st', 'f6su']

    # TODO check this
    # assert cerfa_fields_by_column_name_2009.get('f7vw') is None, "2009 f7vw = {} is not None".format(
    #     cerfa_fields_by_column_name_2009.get('f7vw'))
    assert cerfa_fields_by_column_name_2012.get('f7vw') == ['f7vw']



import itertools
import numpy
import pandas
import pytest

from openfisca_core import periods  # type: ignore
from openfisca_core.tools import assert_near  # type: ignore
from openfisca_france_data.erfs.scenario import ErfsSurveyScenario  # type: ignore
from openfisca_survey_manager.calibration import Calibration  # type: ignore
from openfisca_france.reforms.plf2015 import plf2015  # type: ignore


@pytest.fixture
def fake_calibration(tax_benefit_system):
    def _fake_calibration(fake_input_data: pandas.DataFrame, year: int) -> Calibration:
        input_data_frame = fake_input_data(year)

        survey_scenario = ErfsSurveyScenario.create(
            tax_benefit_system = tax_benefit_system,
            year = year,
            )

        survey_scenario.init_from_data(data = dict(input_data_frame = input_data_frame))

        calibration = Calibration(survey_scenario = survey_scenario, period = year)
        calibration.set_parameters("invlo", 3)
        calibration.set_parameters("up", 3)
        calibration.set_parameters("method", "logit")

        return calibration

    return _fake_calibration


@pytest.mark.skip(reason = "AssertionError: [0. 0.] != [20000. 10000.]")
def test_fake_survey_simulation(tax_benefit_system, fake_input_data, year: int = 2006):
    input_data_frame = fake_input_data(year)
    assert input_data_frame.salaire_imposable.loc[0] == 20000
    assert input_data_frame.salaire_imposable.loc[1] == 10000

    survey_scenario = ErfsSurveyScenario.create(
        tax_benefit_system = tax_benefit_system,
        year = year,
        )

    survey_scenario.init_from_data(data = dict(input_data_frame = input_data_frame))

    assert (input_data_frame.salaire_imposable.loc[0] == 20000).all()
    assert (input_data_frame.salaire_imposable.loc[1] == 10000).all()

    simulation = survey_scenario.simulation

    salaire_imposable = simulation.calculate_add("salaire_imposable", period = year)

    assert (salaire_imposable[0:1] == 20000).all()
    assert (salaire_imposable[1:2] == 10000).all()

    age = simulation.calculate("age", period = periods.period(year).first_month)

    assert age[0] == 77
    assert age[1] == 37

    sal_2003 = simulation.calculate_add("salaire_imposable", period = 2003)
    sal_2004 = simulation.calculate_add("salaire_imposable", period = 2004)
    sal_2005 = simulation.calculate_add("salaire_imposable", period = 2005)
    sal_2006 = simulation.calculate_add("salaire_imposable", period = 2006)

    assert (sal_2003 == 0).all()
    assert (sal_2004 == sal_2006).all()
    assert (sal_2005 == sal_2006).all()

    for year, month in itertools.product(range(2003, 2004), range(1, 13)):
        period = f"{year}-{month}"

        assert (
            simulation
            .calculate_add("salaire_imposable", period = period) == 0
            ).all()

    for year, month in itertools.product(range(2004, 2007), range(1, 13)):
        period = f"{year}-{month}"

        assert (
            simulation
            .calculate("salaire_imposable", period = period) == sal_2006 / 12
            ).all()

    data_frame_by_entity = survey_scenario.create_data_frame_by_entity(
        variables = [
            "age",
            "activite",
            "rsa_base_ressources_i",
            "menage_ordinaire_individus",
            "salaire_imposable",
            "salaire_net",
            "autonomie_financiere",
            "txtppb",
            "af_nbenf",
            "af",
            "af_majoration_enfant",
            "af_age_aine",
            "rsa_base_ressources",
            "rsa",
            "aspa",
            "aide_logement_montant_brut",
            "weight_familles",
            "revenu_disponible",
            "impo",
            ]
        )

    return data_frame_by_entity, simulation


@pytest.mark.skip(
    reason = "AttributeError: 'Calibration' object has no attribute 'simulation'",
    )
def test_fake_calibration_float(fake_calibration, year: int = 2006):
    calibration = fake_calibration(year)
    calibration.total_population = calibration.initial_total_population * 1.123

    revenu_disponible_target = 7e6
    calibration.set_target_margin("revenu_disponible", revenu_disponible_target)
    calibration.calibrate()
    calibration.set_calibrated_weights()

    simulation = calibration.survey_scenario.simulation

    assert_near(
        simulation.calculate("wprm", period=year).sum(),
        calibration.total_population,
        absolute_error_margin = None,
        relative_error_margin = 0.00001,
        )

    assert_near(
        (
            + simulation.calculate("revenu_disponible", period = year)
            * simulation.calculate("wprm", period = year)
            ).sum(),
        revenu_disponible_target,
        absolute_error_margin = None,
        relative_error_margin = 0.00001,
        )


@pytest.mark.skip(
    reason = "ValueError: Length of values does not match length of index",
    )
def test_fake_calibration_age(fake_calibration, fake_input_data, year: int = 2006):
    calibration = fake_calibration(fake_input_data, year)
    survey_scenario = calibration.survey_scenario
    calibration.total_population = calibration.initial_total_population * 1.123
    calibration.set_target_margin("age", [95, 130])
    calibration.calibrate()
    calibration.set_calibrated_weights()

    simulation = survey_scenario.simulation

    assert_near(
        simulation.calculate("wprm").sum(),
        calibration.total_population,
        absolute_error_margin = None,
        relative_error_margin = 0.00001,
        )

    age = (survey_scenario.simulation.calculate("age"),)
    weight_individus = survey_scenario.simulation.calculate("weight_individus")

    for category, target in calibration.margins_by_variable["age"]["target"].items():
        actual = ((age == category) * weight_individus).sum() / weight_individus.sum()
        target = target / numpy.sum(
            calibration
            .margins_by_variable["age"]["target"]
            .values()
            )

        assert_near(
            actual,
            target,
            absolute_error_margin = None,
            relative_error_margin = 0.00001,
            )


def test_calculate_irpp_before_and_after_plf2015(
        tax_benefit_system,
        fake_input_data,
        error_margin: int = 1,
        ):
    # On charge le PLF 2015
    reform = plf2015(tax_benefit_system)

    # Et on va calculer l'IRPP à partir des valeurs du N - 1 (2013)
    year = 2013

    survey_scenario = ErfsSurveyScenario.create(
        tax_benefit_system = reform,
        baseline_tax_benefit_system = tax_benefit_system,
        year = year,
        )

    # On charge les données
    input_data = fake_input_data(2014)

    # On ne garde que les deux parents
    input_data.loc[0, "salaire_imposable"] = 20000
    input_data.loc[1, "salaire_imposable"] = 18000
    input_data = input_data.loc[0:1].copy()

    # On initialise le survey scenario
    survey_scenario.init_from_data(data = dict(input_data_frame = input_data))

    # IRPP avant la réforme
    assert_near(
        survey_scenario.calculate_variable("irpp", use_baseline = True, period = year),
        [-10124, -869],
        error_margin,
        )

    # IRPP après la réforme
    # -911.4 + (1135 - 911.4) = -686.8
    assert_near(
        survey_scenario.calculate_variable("irpp", period = year),
        [-10118, -911.4 + (1135 - 911.4)],
        error_margin,
        )



from functools import reduce

from openfisca_core.reforms import Reform

from openfisca_france.entities import Famille
from openfisca_france_data import france_data_tax_benefit_system

from openfisca_france.reforms import (
    allocations_familiales_imposables,
    cesthra_invalidee,
    # inversion_directe_salaires, # We use a custom inversion_directe_salaires, not merged in the openfisca-france yet
    plf2016,
    plf2016_ayrault_muet,
    plf2015,
    plfr2014,
    trannoy_wasmer,
    )

from openfisca_france_data.reforms import (
    inversion_directe_salaires
    )

# The reforms commented haven't been adapted to the new core API yet.
reform_list = {
    'allocations_familiales_imposables': allocations_familiales_imposables.allocations_familiales_imposables,
    'cesthra_invalidee': cesthra_invalidee.cesthra_invalidee,
    'inversion_directe_salaires': inversion_directe_salaires.inversion_directe_salaires,
    'plf2016': plf2016.plf2016,
    'ayrault_muet': plf2016_ayrault_muet.ayrault_muet,
    'plf2016_counterfactual': plf2016.plf2016_counterfactual,
    'plf2016_counterfactual_2014': plf2016.plf2016_counterfactual_2014,
    'plf2015': plf2015.plf2015,
    'plfr2014': plfr2014.plfr2014,
    'trannoy_wasmer': trannoy_wasmer.trannoy_wasmer,
    }

# Only use the following reform if scipy can be imported
try:
    import scipy
except ImportError:
    scipy = None

if scipy is not None:
    from openfisca_france.reforms import de_net_a_brut
    reform_list['de_net_a_brut'] = de_net_a_brut.de_net_a_brut

reform_by_full_key = {}


def compose_reforms(reforms, tax_benefit_system):
    def compose_reforms_reducer(memo, reform):
        reformed_tbs = reform(memo)
        return reformed_tbs
    final_tbs = reduce(compose_reforms_reducer, reforms, tax_benefit_system)
    return final_tbs


def get_cached_composed_reform(reform_keys, tax_benefit_system):
    full_key = '.'.join(
        [tax_benefit_system.full_key] + reform_keys
        if isinstance(tax_benefit_system, Reform)
        else reform_keys
        )
    composed_reform = reform_by_full_key.get(full_key)

    if composed_reform is None:
        reforms = []
        for reform_key in reform_keys:
            assert reform_key in reform_list, \
                'Error loading cached reform "{}" in build_reform_functions'.format(reform_key)
            reform = reform_list[reform_key]
            reforms.append(reform)
        composed_reform = compose_reforms(
            reforms = reforms,
            tax_benefit_system = tax_benefit_system,
            )
        assert full_key == composed_reform.full_key, (full_key, composed_reform.full_key)
        reform_by_full_key[full_key] = composed_reform
    return composed_reform


def get_cached_reform(reform_key, tax_benefit_system):
    return get_cached_composed_reform([reform_key], tax_benefit_system)




"Calcule divers Smic"

from openfisca_core import periods
from openfisca_france_data import openfisca_france_tax_benefit_system


# smic_horaire_brut = {  # smic horaire brut moyen sur l'année
#     2017: 9.76,
#     2016: 9.67,
#     2015: 9.61,
#     2014: 9.53,
#     2013: 9.43,
#     2012: 9.4 * 6 / 12 + 9.22 * 6 / 12,
#     2011: 9.19 * 1 / 12 + 9.0 * 11 / 12 ,
#     2010: 8.86,
#     }

smic_horaire_brut = dict()
for year in range(2002, 2021):
    try:

        smic_horaire_brut[year] = openfisca_france_tax_benefit_system.parameters(year).marche_travail.salaire_minimum.smic.smic_b_horaire
    except:
        continue

# Sources BDM
# Pour les années avant 2010: BDM donne un indice en base 2000. On multiplie à
# partir du chiffre de 2010. Les résultats pour les années après 2010 sont les
# mêmes à un euro près. 1996-1999 ont la meeeeeême valeur que 2000.
smic_annuel_net_by_year = {
    2022: 4 * 1269.02 + 8 * 1302.64, # latest value as of May, assuming no change over the year to come
    2021: 9 * 1230.6 + 3 * 1258.22,

    2020: 12 * 1218.6,
    2019: 12 * 1204.19,
    2018: 9 * 1173.60 + 3 * 1187.83,  # Baisse de la cotisation chômage en cours d'annnée
    2017: 12 * 1151.50,
    2016: 12 * 1141.61,
    2015: 12 * 1135.99,
    2014: 12 * 1128.70,
    2013: 12 * 1120.43,
    2012: 2 * 1116.87 + 4 * 1118.29 + 6 * 1096.88,
    2011: 11 * 1072.07 + 1094.71,

    2010: 12 * 1056.24,
    2009: 6 * 1050.63 + 6 * 1037.53,
    2008: 6 * 1037.53 + 2 * 1028 + 4 * 1005.36,
    2007: 6 * 1005.36 + 6 * 985.11,
    2006: 6 * 984.61 + 6 * 956.04,
    2005: 12 * 933,
    2004: 12 * 880,
    2003: 12 * 838,
    2002: 12 * 810,
    2001: 12 * 785,
    2000: 12 * 756, # 2000 onwards: based on 151.67 working hours

    1999: 12 * 823,
    1998: 12 * 813,
    1997: 12 * 784,
    1996: 12 * 759,
    1995: 12 * 739,
    1994: 12 * 718,
    1993: 12 * 709,
    1992: 12 * 700,
    1991: 12 * 679,
    1990: 12 * 651,

    1989: 12 * 624,
    1988: 12 * 606,
    1987: 12 * 594,
    1986: 12 * 575,
    1985: 12 * 556,
    1984: 12 * 516,
    1983: 12 * 478,
    1982: 12 * 429, # 1982 onwards: based on 169 working hours
    1981: 12 * 378,
    1980: 12 * 317,

    1979: 12 * 277,
    1978: 12 * 252,
    1977: 12 * 223,
    1976: 12 * 199,
    1975: 12 * 175,
    1974: 12 * 147,
    1973: 12 * 119,
    1972: 12 * 102,
    1971: 12 * 91,
    1970: 12 * 83,

    1969: 12 * 76,
    1968: 12 * 65,
    1967: 12 * 51,
    1966: 12 * 50,
    1965: 12 * 48,
    1964: 12 * 46,
    1963: 12 * 46,
    1962: 12 * 42,
    1961: 12 * 40,
    1960: 12 * 40,

    1959: 12 * 39,
    1958: 12 * 36,
    1957: 12 * 33,
    1956: 12 * 31,
    1955: 12 * 31,
    1954: 12 * 29,
    1953: 12 * 24,
    1952: 12 * 24,
    1951: 12 * 23,
    1950: 12 * 20, # 1950 onwards: based on 173.3 working hours
    }

# get availability of parameters
# p = openfisca_france_tax_benefit_system.parameters

# t = p.prelevements_sociaux.contributions_sociales.csg.activite.imposable.abattement[1]

# t = openfisca_france_tax_benefit_system.parameters.prelevements_sociaux.contributions_sociales.csg.abattement.sous_4pss
# t3 = t.values_list
# sd = t3[t3.__len__() - 1]
# sd.instant_str

# check coverage
start_year = min(smic_annuel_net_by_year.keys())
end_year = max(smic_annuel_net_by_year.keys())

smic_horaire_brut = dict()
for year in range(start_year, end_year+1):
    try:
        # this collects the data from openfisca-france/openfisca_france/parameters/marche_travail/salaire_minimum/smic/smic_b_horaire.yaml ?
        # if year < 1970: log.warning('SMIC before 1970 (SMIG) depends on zone. Which one to use is unclear. TBD.')
        # else : openfisca_france\parameters\marche_travail\salaire_minimum\smic\smic_b_horaire.yaml
        smic_horaire_brut[year] = openfisca_france_tax_benefit_system.get_parameters_at_instant(instant = periods.period(year).start).marche_travail.salaire_minimum.smic.smic_b_horaire
    except:
        continue

# recheck coverage for gross hourly SMIC availability
start_year = min(smic_horaire_brut.keys())
end_year = max(smic_horaire_brut.keys())

def smic_annuel_imposable_from_net(year, smic_hor_brut):
    try:
        # TODO: the formula is not 100 % flexible, I have hard-coded the 4 PSS cut-off; this could be improved in the future
        # then again, it seems not to be used at all for OFF-ERFS, just the smic_horaire_brut
        params = openfisca_france_tax_benefit_system.get_parameters_at_instant(instant = periods.period(year).start)
        smic_net = smic_annuel_net_by_year[year]
        working_hours = params.marche_travail.salaire_minimum.smic.nb_heures_travail_mensuel
        smic_brut = smic_hor_brut * working_hours * 12
        taux_csg = params.prelevements_sociaux.contributions_sociales.csg.activite.imposable.taux
        taux_crds = params.prelevements_sociaux.contributions_sociales.crds.activite.taux
        pss = params.prelevements_sociaux.pss.plafond_securite_sociale_annuel
        abatt_sous_4pss = params.prelevements_sociaux.contributions_sociales.csg.activite.imposable.abattement.rates[0]
        use_plafond = params.prelevements_sociaux.contributions_sociales.csg.activite.imposable.abattement.rates.__len__() == 2
        if use_plafond:
            abatt_dessus_4pss = params.prelevements_sociaux.contributions_sociales.csg.activite.imposable.abattement.rates[1]
        
        # precise formula is kinda unnecessary, since SMIC won't ever be beyond 4 PSS. but still, for the heck of it.
        if use_plafond:
            base_csg_crds = (1 - abatt_sous_4pss) * min(smic_brut, 4 * pss) + (smic_brut > 4 * pss) * (1 - abatt_dessus_4pss) * (smic_brut - 4 * pss)
        else:
            base_csg_crds = (1 - abatt_sous_4pss) * smic_brut

        # final result, add CSG and CRDS to SMIC net
        smic_imposable = (smic_net + (taux_csg + taux_crds) * base_csg_crds)
    except:
        # not all parameters available, return NA
        smic_imposable = None

    return smic_imposable


smic_annuel_imposable_by_year = dict([
    (year, smic_annuel_imposable_from_net(year, smic_horaire_brut[year]))
    for year in range(start_year, end_year)
    ])


smic_horaire_brut_by_year = dict([
    (year, openfisca_france_tax_benefit_system.get_parameters_at_instant(instant = periods.period(year).start).marche_travail.salaire_minimum.smic.smic_b_horaire)
    for year in range(start_year, end_year)
    ])


smic_annuel_brut_by_year = dict([
    (year,
    smic_horaire_brut_by_year[year] * openfisca_france_tax_benefit_system.get_parameters_at_instant(instant = periods.period(year).start).marche_travail.salaire_minimum.smic.nb_heures_travail_mensuel * 12)
    for year in range(start_year, end_year)
    ])


import logging
import numpy as np
import pandas as pd

from openfisca_core import periods
from openfisca_france_data import select_to_match_target
from openfisca_france_data.common import (
    create_salaire_de_base,
    create_traitement_indiciaire_brut,
    create_revenus_remplacement_bruts,
    )
from openfisca_france_data import openfisca_france_tax_benefit_system
from openfisca_france_data.smic import (
    smic_annuel_net_by_year,
    smic_annuel_imposable_by_year,
    )
from openfisca_france_data.utils import assert_dtype
from openfisca_survey_manager.temporary import temporary_store_decorator

log = logging.getLogger(__name__)


@temporary_store_decorator(file_name = 'erfs_fpr')
def build_variables_individuelles(temporary_store = None, year = None):
    """Création des variables individuelles."""

    assert temporary_store is not None
    assert year is not None

    individus = temporary_store['individus_{}_post_01'.format(year)]

    openfisca_by_erfs_variable = {
        'chomage_i': 'chomage_net',
        'pens_alim_recue_i': 'pensions_alimentaires_percues',
        'rag_i': 'rag_net',
        'retraites_i': 'retraite_nette',
        'ric_i': 'ric_net',
        'rnc_i': 'rnc_net',
        'salaires_i': 'salaire_net',
        }

    for variable in openfisca_by_erfs_variable.keys():
        assert variable in individus.columns.tolist(), "La variable {} n'est pas présente".format(variable)

    individus.rename(
        columns = openfisca_by_erfs_variable,
        inplace = True,
        )
    create_variables_individuelles(individus, year)
    assert 'salaire_de_base' in individus.columns , 'salaire de base not in individus'
    temporary_store['individus_{}'.format(year)] = individus
    return individus


# helpers

def create_variables_individuelles(individus, year, survey_year = None):
    """Création des variables individuelles"""

    create_ages(individus, year)
    create_date_naissance(individus, age_variable = None, annee_naissance_variable = 'naia', mois_naissance = 'naim',
         year = year)
    create_activite(individus)
    revenu_type = 'net'
    period = periods.period(year)
    create_revenus(individus, revenu_type = revenu_type)
    create_contrat_de_travail(individus, period = period, salaire_type = revenu_type)
    create_categorie_salarie(individus, period = period, survey_year = survey_year)

    # Il faut que la base d'input se fasse au millésime des données
    # On fait ça car, aussi bien le TaxBenefitSystem et celui réformé peuvent être des réformes
    # Par exemple : si je veux calculer le diff entre le PLF2019 et un ammendement,
    # je besoin d'un droit courant comme même du droit courrant pour l'année des données
    tax_benefit_system = openfisca_france_tax_benefit_system

    # On n'a pas le salaire brut mais le salaire net ou imposable, on doit l'inverser
    create_salaire_de_base(
        individus,
        period = period,
        revenu_type = revenu_type,
        tax_benefit_system = tax_benefit_system
        )

    # Pour les cotisations patronales qui varient avec la taille de l'entreprise'
    create_effectif_entreprise(individus, period = period, survey_year = survey_year)

    # Base pour constituer les familles, foyers, etc.
    create_statut_matrimonial(individus)
    assert 'salaire_de_base' in individus.columns , 'salaire de base not in individus'
    return individus


def create_individu_variables_brutes(individus, revenu_type = None, period = None,
        tax_benefit_system = None, mass_by_categorie_salarie = None,
        calibration_eec = False):
    """
    Crée les variables brutes de revenus:
      - salaire_de_base
      - traitement_indiciaire_brut
      - primes_fonction_publique
      - retraite_bruite
      - chomage_brut
    à partir des valeurs nettes ou imposables de ces revenus
    et d'autres information individuelles
    """
    assert revenu_type in ['imposable', 'net']
    assert period is not None
    assert tax_benefit_system is not None

    assert 'age' in individus.columns

    created_variables = []
    create_contrat_de_travail(individus, period = period, salaire_type = revenu_type)
    created_variables.append('contrat_de_travail')
    created_variables.append('heures_remunerees_volume')

    create_categorie_salarie(individus, period = period)
    created_variables.append('categorie_salarie')
    create_categorie_non_salarie(individus)
    created_variables.append('categorie_non_salarie')

    # FIXME: categorie_non_salarie modifie aussi categorie_salarie  !!
    if (mass_by_categorie_salarie is not None) & (calibration_eec is True):
        calibrate_categorie_salarie(individus, year = None, mass_by_categorie_salarie = mass_by_categorie_salarie)

    create_salaire_de_base(individus, period = period, revenu_type = revenu_type, tax_benefit_system = tax_benefit_system)
    created_variables.append('salaire_de_base')

    create_effectif_entreprise(individus, period = period)
    created_variables.append('effectif_entreprise')

    create_traitement_indiciaire_brut(individus, period = period, revenu_type = revenu_type,
        tax_benefit_system = tax_benefit_system)
    created_variables.append('traitement_indiciaire_brut')
    created_variables.append('primes_fonction_publique')

    create_taux_csg_remplacement(individus, period, tax_benefit_system)
    created_variables.append('taux_csg_remplacement')
    created_variables.append('taux_csg_remplacement_n_1')
    created_variables.append('rfr_special_csg_n')
    created_variables.append('rfr_special_csg_n_1')

    create_revenus_remplacement_bruts(individus, period, tax_benefit_system)
    created_variables.append('chomage_brut')
    created_variables.append('retraite_brute')
    return created_variables


def create_activite(individus):
    """Création de la variable activite.

    0 = Actif occupé
    1 = Chômeur
    2 = Étudiant, élève
    3 = Retraité
    4 = Autre inactif
    """
    create_actrec(individus)

    individus['activite'] = np.nan
    individus.loc[individus.actrec <= 3, 'activite'] = 0
    individus.loc[individus.actrec == 4, 'activite'] = 1
    individus.loc[individus.actrec == 5, 'activite'] = 2
    individus.loc[individus.actrec == 7, 'activite'] = 3
    individus.loc[individus.actrec == 8, 'activite'] = 4
    individus.loc[(individus.age <= 13) | (individus.actrec == 9), 'activite'] = 2
    message = "Valeurs prises par la variable activité \n {}".format(individus['activite'].value_counts(dropna = False))
    assert individus.activite.notnull().all(), message
    assert individus.activite.isin(range(5)).all(), message


def create_actrec(individus):
    """Création de la variables actrec.

    acterc pour activité recodée comme preconisé par l'INSEE p84 du guide méthodologique de l'ERFS
    """
    assert "actrec" not in individus.columns
    individus["actrec"] = np.nan
    acteu = 'act' if 'act' in individus else 'acteu'
    # Attention : Pas de 6, la variable recodée de l'INSEE (voit p84 du guide methodo), ici \
    # la même nomenclature à été adopée
    # 3: contrat a durée déterminée
    individus.loc[individus[acteu]== 1, 'actrec'] = 3
    # 8: femme (homme) au foyer, autre inactif
    individus.loc[individus[acteu] == 3, 'actrec'] = 8
    # 1: actif occupé non salarié
    filter1 = (individus[acteu] == 1) & (individus.stc.isin([1, 3]))  # actifs occupés non salariés à son compte
    individus.loc[filter1, 'actrec'] = 1                             # ou pour un membre de sa famille
    # 2: salarié pour une durée non limitée
    filter2 = (individus[acteu] == 1) & (((individus.stc == 2) & (individus.contra == 1)) | (individus.titc == 2))
    individus.loc[filter2, 'actrec'] = 2
    # 4: au chomage
    filter4 = (individus[acteu] == 2) | ((individus[acteu] == 3) & (individus.mrec == 1))
    individus.loc[filter4, 'actrec'] = 4
    # 5: élève étudiant , stagiaire non rémunéré
    filter5 = (individus[acteu] == 3) & ((individus.forter == 2) | (individus.rstg == 1))
    individus.loc[filter5, 'actrec'] = 5
    # 7: retraité, préretraité, retiré des affaires unchecked
    try: # cas >= 2014, evite de ramener l'année dans la fonction
        filter7 = (individus[acteu] == 3) & ((individus.ret == 1))
    except Exception:
        pass
    try: # cas 2004 - 2013
        filter7 = (individus[acteu] == 3) & ((individus.retrai == 1) | (individus.retrai == 2))
    except Exception: # cas 1996 - 2003
        cstot = 'dcstot' if 'dcstot' in individus else 'cstotr'
        filter7 = (individus[acteu] == 3) & ((individus[cstot] == 7))

    individus.loc[filter7, 'actrec'] = 7
    # 9: probablement enfants de - de 16 ans TODO: check that fact in database and questionnaire
    individus.loc[individus[acteu] == 0, 'actrec'] = 9

    assert individus.actrec.notnull().all()
    individus.actrec = individus.actrec.astype("int8")
    assert_dtype(individus.actrec, "int8")

    assert (individus.actrec != 6).all(), 'actrec ne peut pas être égale à 6'
    assert individus.actrec.isin(range(1, 10)).all(), 'actrec values are outside the interval [1, 9]'


def create_ages(individus, year = None):
    """Création des variables age et age_en_moi."""
    assert year is not None
    individus['age'] = year - individus.naia - 1
    individus['age_en_mois'] = 12 * individus.age + 12 - individus.naim  # TODO why 12 - naim

    for variable in ['age', 'age_en_mois']:
        assert individus[variable].notnull().all(), "Il y a {} entrées non renseignées pour la variable {}".format(
            individus[variable].notnull().sum(), variable)


def create_categorie_salarie(individus, period, survey_year = None):
    """Création de la variable categorie_salarie.

    Ses modalités sont;
      - "prive_non_cadre
      - "prive_cadre
      - "public_titulaire_etat
      - "public_titulaire_militaire
      - "public_titulaire_territoriale
      - "public_titulaire_hospitaliere
      - "public_non_titulaire
      - "non_pertinent"
    à partir des variables de l'eec' :
      - chpub :
          Variable déjà présente dans les fichiers antérieurs à 2013. Cependant, ses modalités ont été réordonnées ainsi :
          les modalités 1, 2, 3, 4, 5 et 6 de la variable CHPUB dans les fichiers antérieurs à 2013 correspondent à partir de
          2013 aux modalités respectives 3, 4, 5, 7, 2 et 1. Par ailleurs, une nouvelle modalité, Sécurité sociale (6) a été
          créée en 2013.
          A partir de (>=) 2013:
              1 Entreprise privée ou association
              2 Entreprise publique (EDF, La Poste, SNCF, etc.)
              3 État
              4 Collectivités territoriales
              5 Hôpitaux publics
              6 Sécurité sociale
              7 Particulier
          Avant (<) 2013:
              1 - Etat
              2 - Collectivités locales, HLM
              3 - Hôpitaux publics
              4 - Particulier
              5 - Entreprise publique (La Poste, EDF-GDF, etc.)
              6 - Entreprise privée, association

      - encadr : (encadrement de personnes)
          1 - Oui
          2 - Non

      - prosa (survey_year < 2013, voir qprc)
          1 Manoeuvre ou ouvrier spécialisé
          2 Ouvrier qualifié ou hautement qualifié
          3 Technicien
          4 Employé de bureau, de commerce, personnel de services, personnel de catégorie C ou D
          5 Agent de maîtrise, maîtrise administrative ou commerciale VRP (non cadre) personnel de catégorie B
          7 Ingénieur, cadre (à l'exception des directeurs généraux ou de ses adjoints directs) personnel de catégorie A
          8 Directeur général, adjoint direct
          9 Autre

      - qprc (survey_year >= 2013, voir prosa)
          Vide Sans objet (personnes non actives occupées, travailleurs informels et travailleurs intérimaires,
                          en activité temporaire ou d'appoint)
          1 Manoeuvre ou ouvrier spécialisé
          2 Ouvrier qualifié ou hautement qualifié, technicien d'atelier
          3 Technicien
          4 Employé de bureau, employé de commerce, personnel de services, personnel de catégorie C
          dans la fonction publique
          5 Agent de maîtrise, maîtrise administrative ou commerciale, VRP (non cadre), personnel de
          catégorie B dans la fonction publique
          6 Ingénieur, cadre (à l'exception des directeurs ou de leurs adjoints directs) personnel de
          catégorie A dans la fonction publique
          7 Directeur général, adjoint direct
          8 Autre
          9 Non renseigné

      - statut :
          11 - Indépendants
          12 - Employeurs
          13 - Aides familiaux
          21 - Intérimaires
          22 - Apprentis
          33 - CDD (hors Etat, coll.loc.), hors contrats aides
          34 - Stagiaires et contrats aides (hors Etat, coll.loc.)
          35 - Autres contrats (hors Etat, coll.loc.)
          43 - CDD (Etat, coll.loc.), hors contrats aides
          44 - Stagiaires et contrats aides (Etat, coll.loc.)
          45 - Autres contrats (Etat, coll.loc.)

      - titc :
          1 - Elève fonctionnaire ou stagiaire
          2 - Agent titulaire
          3 - Contractuel

    """
    assert period is not None
    if survey_year is None:
        survey_year = period.start.year

    if survey_year >= 2013:
        log.debug(f"Using qprcent to infer chpub for year {survey_year}")
        chpub_replacement = {
            0: 0,
            3: 1,
            4: 2,
            5: 3,
            7: 4,
            2: 5,
            1: 6,
            6: 1,
            }
        individus['chpub'] = individus.chpub.map(chpub_replacement)

        log.debug('Using qprc to infer prosa for year {}'.format(survey_year))
        qprc_to_prosa = {
            0: 0,
            1: 1,
            2: 2,
            3: 3,
            4: 4,
            5: 5,
            6: 7,
            7: 8,
            8: 9,
            9: 5,  # On met les non renseignés en catégorie B
            }
        individus['prosa'] = individus.qprcent.map(qprc_to_prosa)
        # Actually prosa is there I don't need to change it further
    else:
        pass

    assert individus.chpub.isin(range(0, 7)).all(), \
        "chpub n'est pas toujours dans l'intervalle [0, 6]\n{}".format(individus.chpub.value_counts(dropna = False))

# N'existe qu'à partir de 2006 inclu
# On pourrait prendre csei mais elle ne recoupe pas bien prosa, ce serait vraiment
# une autre manière de déterminer la catégorie salariale.
    if 'encadr' in individus:
       individus.loc[individus.encadr == 0, 'encadr'] = 2
       assert individus.encadr.isin(range(1, 3)).all(), \
           "encadr n'est pas toujours dans l'intervalle [1, 2]\n{}".format(individus.encadr.value_counts(dropna = False))

    assert individus.prosa.isin(range(0, 10)).all(), \
        "prosa n'est pas toujours dans l'intervalle [0, 9]\n{}".format(individus.prosa.value_counts())

    statut_values = [0, 11, 12, 13, 21, 22, 33, 34, 35, 43, 44, 45, 99]
    assert individus.statut.isin(statut_values).all(), \
        "statut n'est pas toujours dans l'ensemble {} des valeurs antendues.\n{}".format(
            statut_values,
            individus.statut.value_counts(dropna = False)
            )

    if survey_year >= 2013:
        if individus.titc.isnull().any():
            individus.loc[
                individus.titc.isnull() & ~(individus.chpub.isin([3, 4, 5])),
                'titc'
            ] = 0
        assert individus.titc.isin(range(5)).all(), \
            "titc n'est pas toujours dans l'ensemble [0, 1, 2, 3, 4] des valeurs antendues.\n{}".format(
                individus.titc.value_counts(dropna = False)
                )
    else:
        assert individus.titc.isin(range(4)).all(), \
            "titc n'est pas toujours dans l'ensemble [0, 1, 2, 3] des valeurs antendues.\n{}".format(
                individus.titc.value_counts(dropna = False)
                )

    chpub = individus.chpub
    titc = individus.titc

    # encadrement
    assert 'cadre' not in individus.columns
    individus['cadre'] = False
    individus.loc[individus.prosa.isin([7, 8]), 'cadre'] = True
    if 'encadr' in individus: #N'existe qu'à partir de 2006 inclu
        individus.loc[(individus.prosa == 9) & (individus.encadr == 1), 'cadre'] = True
    cadre = (
        (individus.statut >= 21) & (individus.statut <= 35)  # En activité hors fonction publique
        & (chpub > 3) # Hors fonction publique mais entreprise publique
        & individus.cadre
        )
    del individus['cadre']

    # etat_stag = (chpub == 1) & (titc == 1)
    etat_titulaire = (chpub == 1) & ((titc == 2) | (titc == 1))
    etat_contractuel = (chpub == 1) & (titc == 3)

    militaire = False  # TODO:

    # collect_stag = (chpub==2) & (titc == 1)
    collectivites_locales_titulaire = (chpub == 2) & ((titc == 2) | (titc == 1))
    collectivites_locales_contractuel = (chpub == 2) & (titc == 3)

    # hosp_stag = (chpub==2)*(titc == 1)
    hopital_titulaire = (chpub == 3) & ((titc == 2) | (titc == 1))
    hopital_contractuel = (chpub == 3) & (titc == 3)
    contractuel = collectivites_locales_contractuel | hopital_contractuel | etat_contractuel

    if 'salaire_net' in individus and (individus['salaire_net'] > 0).any():
        actif_occupe = individus['salaire_net'] > 0
    elif 'salaire_imposable' in individus and (individus['salaire_imposable'] > 0).any():
        actif_occupe = individus['salaire_imposable'] > 0
    else:
        assert False, u'Pas de variable de salaire disponible non nuelle pour déterminer si un actif est occupé'

    #   TODO: may use something like this but to be improved and tested
    #    individus['categorie_salarie'] = np.select(
    #         [non_cadre, cadre, etat_titulaire, militaire, collectivites_locales_titulaire, hopital_titulaire,
    #               contractuel, non_pertinent]  # Choice list
    #         [0, 1, 2, 3, 4, 5, 6, 7],  # Condlist
    #         )

    individus['categorie_salarie'] = actif_occupe * (
        0 +
        1 * cadre +
        2 * etat_titulaire +
        3 * militaire +
        4 * collectivites_locales_titulaire +
        5 * hopital_titulaire +
        6 * contractuel
        ) + np.logical_not(actif_occupe) * 7

    log.debug('Les valeurs de categorie_salarie sont: \n {}'.format(
        individus['categorie_salarie'].value_counts(dropna = False)))
    assert individus['categorie_salarie'].isin(range(8)).all(), \
        "categorie_salarie n'est pas toujours dans l'intervalle [0, 7]\n{}".format(
            individus.categorie_salarie.value_counts(dropna = False))


def create_categorie_non_salarie(individus):
    """Création de la variable categorie_salarie.

    Ses modalités sont:
      - "non_pertinent
      - "artisan
      - "commercant
      - "profession_liberale
    à partir des variables de l'eec' :
      - cstot
        - 00 Non renseigné (pour les actifs)
        - 11 Agriculteurs sur petite exploitation
        - 12 Agriculteurs sur moyenne exploitation
        - 13 Agriculteurs sur grande exploitation
        - 21 Artisans
        - 22 Commerçants et assimilés
        - 23 Chefs d'entreprise de 10 salariés ou plus
        - 31 Professions libérales
        - 33 Cadres de la fonction publique
        - 34 Professeurs, professions scientifiques
        - 35 Professions de l'information, des arts et des spectacles
        - 37 Cadres administratifs et commerciaux d'entreprise
        - 38 Ingénieurs et cadres techniques d'entreprise
        - 42 Professeurs des écoles, instituteurs et assimilés
        - 43 Professions intermédiaires de la santé et du travail social
        - 44 Clergé, religieux
        - 45 Professions intermédiaires administratives de la fonction publique
        - 46 Professions intermédiaires administratives et commerciales des entreprises
        - 47 Techniciens
        - 48 Contremaîtres, agents de maîtrise
        - 52 Employés civils et agents de service de la fonction publique
        - 53 Policiers et militaires
        - 54 Employés administratifs d'entreprise
        - 55 Employés de commerce
        - 56 Personnels des services directs aux particuliers
        - 62 Ouvriers qualifiés de type industriel
        - 63 Ouvriers qualifiés de type artisanal
        - 64 Chauffeurs
        - 65 Ouvriers qualifiés de la manutention, du magasinage et du transport
        - 67 Ouvriers non qualifiés de type industriel
        - 68 Ouvriers non qualifiés de type artisanal
        - 69 Ouvriers agricoles
        - 71 Anciens agriculteurs exploitants
        - 72 Anciens artisans, commerçants, chefs d'entreprise
        - 74 Anciens cadres
        - 75 Anciennes professions intermédiaires
        - 77 Anciens employés
        - 78 Anciens ouvriers
        - 81 Chômeurs n'ayant jamais travaillé
        - 83 Militaires du contingent
        - 84 Elèves, étudiants
        - 85 Personnes diverses sans activité professionnelle de moins de 60 ans (sauf retraités)
        - 86 Personnes diverses sans activité professionnelle de 60 ans et plus (sauf retraités)
    """
    assert individus.cstot.notnull().all()
    if not pd.api.types.is_numeric_dtype(individus.cstot):
        individus.replace(
            {
                'cstot' : {'': '0', '00': '0'}
                },
            inplace = True
            )

    individus['cstot'] = individus.cstot.astype('int')
    assert set(individus.cstot.unique()) < set([
        0,
        11, 12, 13,
        21, 22, 23,
        31, 33, 34, 35, 37, 38,
        42, 43, 44, 45, 46, 47, 48,
        52, 53, 54, 55, 56,
        62, 63, 64, 65, 67, 68, 69,
        71, 72, 74, 75, 77, 78,
        81, 83, 84, 85, 86,
        ])

    agriculteur = individus.cstot.isin([11, 12, 13])
    artisan = individus.cstot.isin([21])
    commercant = individus.cstot.isin([22])
    chef_entreprise = individus.cstot.isin([23])
    profession_liberale = individus.cstot.isin([31])
    individus.loc[
        agriculteur | artisan,
        'categorie_non_salarie'
        ] = 1
    individus.loc[
        commercant | chef_entreprise,
        'categorie_non_salarie'
        ] = 2
    individus.loc[
        profession_liberale,
        'categorie_non_salarie'
        ] = 3
    # Correction fonction publique
    individus.loc[
        (
            (individus.categorie_salarie == 0)
            & (individus.cstot.isin([31, 33, 34, 35, 37, 38,]))
            ),
        'categorie_salarie'
        ] = 1

    # Correction encadrement
    individus.loc[
        (
            (individus.categorie_salarie == 0)
            & (individus.cstot.isin([31, 34, 35, 37 ,38]))  # Cadres hors FP
            ),
        'categorie_salarie'
        ] = 1
    # Correction fonction publique
    individus.loc[
        (
            (individus.categorie_salarie.isin([0, 1, 7]))
            & (individus.cstot == 53) # Policiers et militaires reversé dans titulaire état
            ),
        'categorie_salarie'
        ] = 2


def create_contrat_de_travail(individus, period, salaire_type = 'imposable'):
    """Création de la variable contrat_de_travail et heure_remunerees_volume.

    Ses modliatés sont:
        0 - temps_plein
        1 - temps_partiel
        2 - forfait_heures_semaines
        3 - forfait_heures_mois
        4 - forfait_heures_annee
        5 - forfait_jours_annee
        6 - sans_objet
    à partir de la variables tppred (Temps de travail dans l'emploi principal)
        0 - Sans objet (ACTOP='2') ou non renseigné principal
        1 - Temps complet
        2 - Temps partiel
    qui doit être construite à partir de tpp
        0 - Sans objet (ACTOP='2') ou non renseigné
        1 - A temps complet
        2 - A temps partiel
        3 - Sans objet (pour les personnes non salariées qui estiment que cette question ne
            s'applique pas à elles)
    et de duhab
        1 - Temps partiel de moins de 15 heures
        2 - Temps partiel de 15 à 29 heures
        3 - Temps partiel de 30 heures ou plus
        4 - Temps complet de moins de 30 heures
        5 - Temps complet de 30 à 34 heures
        6 - Temps complet de 35 à 39 heures
        7 - Temps complet de 40 heures ou plus
        9 - Pas d'horaire habituel ou horaire habituel non déclaré
    On utilise également hcc pour déterminer le nombre d'heures
    TODO: utiliser la variable forfait
    """
    if not isinstance(period, periods.Period):
        period = periods.period(period)

    assert salaire_type in ['net', 'imposable']

    if ((individus.hhc.dtype != 'float') & (individus.hhc.dtype != 'float32')):
        individus.loc[individus.hhc == "", "hhc"] = np.nan
        individus.hhc = individus.hhc.astype(float)
    individus.loc[individus.hhc <= 0.01 , "hhc"] = np.nan

    assert ((individus.hhc > 0) | individus.hhc.isnull()).all()

    assert individus.tppred.isin(range(3)).all(), \
        'tppred values {} should be in [0, 1, 2]'.format(individus.tppred.unique())

    assert (
        individus.duhab.isin(range(10)) & (individus.duhab != 8)
        ).all(), 'duhab values {} should be in [0, 1, 2, 3, 4, 5, 6, 7, 9]'.format(individus.duhab.unique())

    individus['contrat_de_travail'] = 6  # sans objet par défaut
    if salaire_type == 'net':
        assert (individus.query('salaire_net == 0').contrat_de_travail == 6).all()
        log.debug('Salaire retenu: {}'.format('salaire_net'))
        individus['salaire'] = individus.salaire_net.copy()
        smic = smic_annuel_net_by_year[period.start.year]

    elif salaire_type == 'imposable':
        individus['salaire_imposable'] = individus.salaire_imposable.fillna(0)
        assert (individus.query('salaire_imposable == 0').contrat_de_travail == 6).all()
        log.debug('Salaire retenu: {}'.format('salaire_imposable'))
        individus['salaire'] = individus.salaire_imposable.copy()
        smic = smic_annuel_imposable_by_year[period.start.year]

    if period.unit == 'month':
        smic = period.size * smic / 12

    # 0 On élimine les individus avec un salaire_net nul des salariés
    # 1. utilisation de tppred et durhab
    # 1.1  temps_plein
    individus.query('tppred == 1').duhab.value_counts(dropna = False)
    assert (individus.query('tppred == 1').duhab >= 4).all()
    individus.loc[
        (individus.salaire > 0) & (
            (individus.tppred == 1) | (individus.duhab.isin(range(4, 8)))
            ),
        'contrat_de_travail'
        ] = 0
    assert (individus.query('salaire == 0').contrat_de_travail == 6).all()
    # 1.2 temps partiel
    assert (individus.query('tppred == 2').duhab.isin([1, 2, 3, 9])).all()
    individus.loc[
        (individus.salaire > 0) & (
            (individus.tppred == 2) | (individus.duhab.isin(range(1, 4)))
            ),
        'contrat_de_travail'
        ] = 1
    assert (individus.query('salaire == 0').contrat_de_travail == 6).all()
    # 2. On traite les salaires horaires inféreurs au SMIC
    # 2.1 temps plein
    temps_plein = individus.query('(contrat_de_travail == 0) & (salaire > 0)')
    # (temps_plein.salaire > smic).value_counts()
    # temps_plein.query('salaire < 15000').salaire.hist()
    individus['heures_remunerees_volume'] = 0
    # On bascule à temps partiel et on réajuste les heures des temps plein qui ne touche pas le smic
    temps_plein_sous_smic = (
        (individus.contrat_de_travail == 0) &
        (individus.salaire > 0) &
        (individus.salaire < smic)
        )
    individus.loc[
        temps_plein_sous_smic,
        'contrat_de_travail'] = 1
    individus.loc[
        temps_plein_sous_smic,
        'heures_remunerees_volume'] = individus.loc[
            temps_plein_sous_smic,
            'salaire'
            ] / smic * 35
    assert (individus.loc[temps_plein_sous_smic, 'heures_remunerees_volume'] < 35).all()
    assert (individus.loc[temps_plein_sous_smic, 'heures_remunerees_volume'] > 0).all()
    assert (individus.query('salaire == 0').contrat_de_travail == 6).all()
    del temps_plein, temps_plein_sous_smic
    # 2.2 Pour les temps partiel on prends les heures hhc
    # On vérfie que celles qu'on a créées jusqu'ici sont correctes
    assert (individus.query('salaire == 0').contrat_de_travail == 6).all()
    assert (individus.query('(contrat_de_travail == 1) & (salaire > 0)').heures_remunerees_volume < 35).all()
    #
    axes = (individus.query('(contrat_de_travail == 1) & (salaire > 0)').hhc).hist(bins=100)
    axes.set_title("Heures (hhc)")
    # individus.query('(contrat_de_travail == 1) & (salaire > 0)').hhc.isnull().sum() = 489
    # 2.2.1 On abaisse le nombre d'heures pour que les gens touchent au moins le smic horaire
    temps_partiel = (individus.contrat_de_travail == 1) & (individus.salaire > 0)
    moins_que_smic_horaire_hhc = (
        ((individus.salaire / individus.hhc) < (smic / 35)) &
        individus.hhc.notnull()
        )
    # Si on dispose de la variable hhc on l'utilise
    individus.loc[
        temps_partiel & moins_que_smic_horaire_hhc,
        'heures_remunerees_volume'
        ] = individus.loc[
            temps_partiel & moins_que_smic_horaire_hhc,
            'salaire'
            ] / smic * 35
    individus.loc[
        temps_partiel & (~moins_que_smic_horaire_hhc) & individus.hhc.notnull(),
        'heures_remunerees_volume'
        ] = individus.loc[
            temps_partiel & (~moins_que_smic_horaire_hhc) & individus.hhc.notnull(),
            'hhc'
            ]
    axes = (
        individus
        .loc[temps_partiel]
        .query('(contrat_de_travail == 1) & (salaire > 0)')
        .heures_remunerees_volume
        .hist(bins=100)
        )
    axes.set_title("Heures (heures_remunerees_volume)")
    # 2.2.2 Il reste à ajuster le nombre d'heures pour les salariés à temps partiel qui n'ont pas de hhc
    # et qui disposent de moins que le smic_horaire ou de les basculer en temps plein sinon
    moins_que_smic_horaire_sans_hhc = (individus.salaire < smic) & individus.hhc.isnull()
    individus.loc[
        temps_partiel & moins_que_smic_horaire_sans_hhc,
        'heures_remunerees_volume'
        ] = individus.loc[
            temps_partiel & moins_que_smic_horaire_sans_hhc,
            'salaire'
            ] / smic * 35
    plus_que_smic_horaire_sans_hhc = (
        (individus.salaire >= smic) &
        individus.hhc.isnull()
        )
    individus.loc[
        temps_partiel & plus_que_smic_horaire_sans_hhc,
        'contrat_de_travail'
        ] = 0
    individus.loc[
        temps_partiel & plus_que_smic_horaire_sans_hhc,
        'heures_remunerees_volume'
        ] = 0
    # 2.2.3 On repasse en temps plein ceux qui ont plus de 35 heures par semaine
    temps_partiel_bascule_temps_plein = (
        temps_partiel &
        (individus.heures_remunerees_volume >= 35)
        )
    individus.loc[
        temps_partiel_bascule_temps_plein,
        'contrat_de_travail',
        ] = 0
    individus.loc[
        temps_partiel_bascule_temps_plein,
        'heures_remunerees_volume',
        ] = 0
    del temps_partiel, temps_partiel_bascule_temps_plein, moins_que_smic_horaire_hhc, moins_que_smic_horaire_sans_hhc
    assert (individus.query('contrat_de_travail == 0').heures_remunerees_volume == 0).all()
    assert (individus.query('contrat_de_travail == 1').heures_remunerees_volume < 35).all()
    assert (individus.query('salaire == 0').contrat_de_travail == 6).all()
    assert (individus.query('contrat_de_travail == 6').heures_remunerees_volume == 0).all()
    # 2.3 On traite ceux qui ont un salaire mais pas de contrat de travail renseigné
    # (temps plein ou temps complet)
    salarie_sans_contrat_de_travail = (
        (individus.salaire > 0) &
        ~individus.contrat_de_travail.isin([0, 1])
        )
    # 2.3.1 On passe à temps plein ceux qui ont un salaire supérieur au SMIC annuel
    individus.loc[
        salarie_sans_contrat_de_travail &
        (individus.salaire >= smic),
        'contrat_de_travail'
        ] = 0
    assert (individus.loc[
        salarie_sans_contrat_de_travail &
        (individus.salaire >= smic),
        'heures_remunerees_volume'
        ] == 0).all()
    # 2.3.2 On passe à temps partiel ceux qui ont un salaire inférieur au SMIC annuel
    individus.loc[
        salarie_sans_contrat_de_travail &
        (individus.salaire < smic),
        'contrat_de_travail'
        ] = 1
    individus.loc[
        salarie_sans_contrat_de_travail &
        (individus.salaire < smic),
        'heures_remunerees_volume'
        ] = individus.loc[
            salarie_sans_contrat_de_travail &
            (individus.salaire < smic),
            'salaire'
            ] / smic * 35
    # 2.3.3 On attribue des heures rémunérées aux individus à temps partiel qui ont un
    # salaire strictement positif mais un nombre d'heures travaillées nul
    salaire_sans_heures = (individus.contrat_de_travail == 1) & ~(individus.heures_remunerees_volume > 0)
    assert (individus.loc[salaire_sans_heures, 'salaire'] > 0).all()
    assert (individus.loc[salaire_sans_heures, 'duhab'] == 1).all()
    # Cela concerne peu de personnes qui ont par ailleurs duhab = 1 et un salaire supérieur au smic.
    # On leur attribue donc un nombre d'heures travaillées égal à 15.
    individus.loc[
        salaire_sans_heures &
        (individus.salaire > smic),
        'heures_remunerees_volume'] = 15
    #
    individus.query('salaire > 0').contrat_de_travail.value_counts(dropna = False)
    individus.query('salaire == 0').contrat_de_travail.value_counts(dropna = False)

    individus.loc[salarie_sans_contrat_de_travail, 'salaire'].min()
    individus.loc[salarie_sans_contrat_de_travail, 'salaire'].hist(bins = 1000)
    del salarie_sans_contrat_de_travail
    # On vérifie que l'on n'a pas fait d'erreurs
    assert (individus.salaire >= 0).all(), "Des salaires sont negatifs: {}".format(
            individus.loc[~(individus.salaire >= 0), 'salaire']
            )
    assert individus.contrat_de_travail.isin([0, 1, 6]).all()
    assert (individus.query('salaire > 0').contrat_de_travail.isin([0, 1])).all()
    assert (individus.query('salaire == 0').contrat_de_travail == 6).all()
    assert (individus.query('salaire == 0').heures_remunerees_volume == 0).all()
    assert (individus.query('contrat_de_travail in [0, 6]').heures_remunerees_volume == 0).all()
    assert (individus.query('contrat_de_travail == 1').heures_remunerees_volume < 35).all()
    assert (individus.heures_remunerees_volume >= 0).all()
    assert (individus.query('contrat_de_travail == 1').heures_remunerees_volume > 0).all(), \
        "Des heures des temps partiels ne sont pas strictement positives: {}".format(
            individus.query('contrat_de_travail == 1').loc[
                ~(individus.query('contrat_de_travail == 1').heures_remunerees_volume > 0),
                'heures_remunerees_volume'
                ]
            )

    # la variable heures_remunerees_volume est nombre d'heures par semaine. On ajuste selon la période
    period = periods.period(period)
    if period.unit == 'year':
        individus['heures_remunerees_volume'] = individus.heures_remunerees_volume * 52
    elif period.unit == 'month':
        if period.size == 1:
            individus['heures_remunerees_volume'] = individus.heures_remunerees_volume * 52 / 12
        elif period.size == 3:
            individus['heures_remunerees_volume'] = individus.heures_remunerees_volume * 52 / 4
        else:
            log.error('Wrong period {}. Should be month or year'.format(period))
            raise
    else:
        log.error('Wrong period {}. Should be month or year'.format(period))
        raise

    del individus['salaire']

    return


def create_date_naissance(individus, age_variable = 'age', annee_naissance_variable = None, mois_naissance = None,
         year = None):
    """
    Création de la variable date_naissance à partir des variables age ou de l'année et du mois de naissance
    """
    assert year is not None
    assert bool(age_variable) != bool(annee_naissance_variable)  # xor

    random_state = np.random.RandomState(42)
    month_birth = 1 + random_state.randint(12, size = len(individus))
    day_birth = 1 + random_state.randint(28, size = len(individus))

    if age_variable is not None:
        assert age_variable in individus
        year_birth = (year - individus[age_variable]).astype(int)

    elif annee_naissance_variable is not None:
        year_birth = individus[annee_naissance_variable].astype(int)
        if mois_naissance is not None:
            month_birth = individus[mois_naissance].astype(int)

    # sometimes age is actually the year, fixing this by adding back year
    year_birth[year_birth < 1000] += year

    individus['date_naissance'] = pd.to_datetime(
        pd.DataFrame({
            'year': year_birth,
            'month': month_birth,
            'day': day_birth,
            })
        )


def create_effectif_entreprise(individus, period = None, survey_year = None):
    """Création de la variable effectif_entreprise.

    A partir de la variable nbsala qui prend les valeurs suivantes:
    Création de la variable effectif_entreprise à partir de la variable nbsala qui prend les valeurs suivantes:
    A partir de (>=) 2013
        0 Vide Sans objet ou non renseigné
        1 1 salarié
        2 2 salariés
        3 3 salariés
        4 4 salariés
        5 5 salariés
        6 6 salariés
        7 7 salariés
        8 8 salariés
        9 9 salariés
        10 10 à 49 salariés
        11 50 à 499 salariés
        12 500 salariés ou plus
        99 Ne sait pas
    Strictement avant (<) 2013
        0 - Non pertinent
        1 - Aucun salarié
        2 - 1 ou 4 salariés
        3 - 5 à 9 salariés
        4 - 10 à 19 salariés
        5 - 20 à 49 salariés
        6 - 50 à 199 salariés
        7 - 200 à 499 salariés
        8 - 500 à 999 salariés
        9 - 1000 salariés ou plus
        99 - Ne sait pas
    """
    assert period is not None
    if survey_year is None:
        survey_year = period.start.year

    if survey_year >= 2013:
        assert individus.nbsala.isin(list(range(0, 13)) + [99]).all(), \
            "nbsala n'est pas toujours dans l'intervalle [0, 12] ou 99 \n{}".format(
                individus.nbsala.value_counts(dropna = False))
        individus['effectif_entreprise'] = np.select( # condition_lits, choice_list
            [
                individus.nbsala == 0,  # 0
                individus.nbsala == 1,  # 1
                individus.nbsala == 2,  # 2
                individus.nbsala == 3,  # 3
                individus.nbsala == 4,  # 4
                individus.nbsala == 5,  # 5
                individus.nbsala == 6,  # 6
                individus.nbsala == 7,  # 7
                individus.nbsala == 8,  # 8
                individus.nbsala == 9,  # 9
                individus.nbsala == 10,  # 9
                (individus.nbsala == 11) | (individus.nbsala == 99),  # 9
                individus.nbsala == 12,  # 9
                ],
            [0, 1, 2, 3, 4, 5, 5, 7, 8, 9, 10, 50, 500],
            )
        assert individus.effectif_entreprise.isin([0, 1, 2, 3, 4, 5, 5, 7, 8, 9, 10, 50, 500]).all(), \
            "effectif_entreprise n'est pas toujours dans [0, 1, 5, 10, 20, 50, 200, 500, 1000] \n{}".format(
                individus.effectif_entreprise.value_counts(dropna = False))

    else:
        assert individus.nbsala.isin(list(range(0, 10)) + [99]).all(), \
            "nbsala n'est pas toujours dans l'intervalle [0, 9] ou 99 \n{}".format(
                individus.nbsala.value_counts(dropna = False))
        individus['effectif_entreprise'] = np.select(
            [
                individus.nbsala.isin([0, 1]),  # 0
                individus.nbsala == 2,  # 1
                individus.nbsala == 3,  # 5
                individus.nbsala == 4,  # 10
                individus.nbsala == 5,  # 20
                (individus.nbsala == 6) | (individus.nbsala == 99),  # 50
                individus.nbsala == 7,  # 200
                individus.nbsala == 8,  # 500
                individus.nbsala == 9,  # 1000
                ],
            [0, 1, 5, 10, 20, 50, 200, 500, 1000],
            )

        assert individus.effectif_entreprise.isin([0, 1, 5, 10, 20, 50, 200, 500, 1000]).all(), \
            "effectif_entreprise n'est pas toujours dans [0, 1, 5, 10, 20, 50, 200, 500, 1000] \n{}".format(
                individus.effectif_entreprise.value_counts(dropna = False))
        log.debug('Effectif entreprise:\n{}'.format(
            individus.effectif_entreprise.value_counts(dropna = False)))


def create_revenus(individus, revenu_type = 'imposable'):
    """Création des plusieurs variablesde revenu.

    Ces variables sont:
        chomage_net,
        pensions_alimentaires_percues,
        rag_net,
        retraite_nette,
        ric_net,
        rnc_net,
    et éventuellement, si revenus_type = 'imposable' des variables:
        chomage_imposable,
        rag,
        retraite_imposable,
        ric,
        rnc,
        salaire_imposable,
    """
    individus['chomage_brut'] = individus.csgchod_i + individus.chomage_net
    individus['retraite_brute'] = individus.csgrstd_i + individus.retraite_nette

    if revenu_type == 'imposable':
        variables = [
            # 'pension_alimentaires_percues',
            'chomage_imposable',
            'retraite_imposable',
            ]
        for variable in variables:
            assert variable in individus.columns.tolist(), "La variable {} n'est pas présente".format(variable)

        for variable in variables:
            if (individus[variable] < 0).any():

                negatives_values = individus[variable].value_counts().loc[individus[variable].value_counts().index < 0]
                log.debug("La variable {} contient {} valeurs négatives\n {}".format(
                    variable,
                    negatives_values.sum(),
                    negatives_values,
                    )
                )

    # csg des revenus de replacement
    # 0 - Non renseigné/non pertinent
    # 1 - Exonéré
    # 2 - Taux réduit
    # 3 - Taux plein
    taux = pd.concat(
        [
            individus.csgrstd_i / individus.retraite_brute,
            individus.csgchod_i / individus.chomage_brut,
            ],
        axis = 1
        ).max(axis = 1)

    # taux.loc[(0 < taux) & (taux < .1)].hist(bins = 100)
    individus['taux_csg_remplacement'] = np.select(
        [
            taux.isnull(),
            taux.notnull() & (taux < 0.021),
            taux.notnull() & (taux > 0.021) & (taux < 0.0407),
            taux.notnull() & (taux > 0.0407)
            ],
        [0, 1, 2, 3]
        )
    for value in [0, 1, 2, 3]:
        assert (individus.taux_csg_remplacement == value).any(), \
            "taux_csg_remplacement ne prend jamais la valeur {}".format(value)
    assert individus.taux_csg_remplacement.isin(range(4)).all()


def create_statut_matrimonial(individus):
    """
    Création de la variable statut_marital qui prend les valeurs:
      1 - "Marié",
      2 - "Célibataire",
      3 - "Divorcé",
      4 - "Veuf",
      5 - "Pacsé",
      6 - "Jeune veuf"
    à partir de la variable matri qui prend les valeurs
      0 - sans objet (moins de 15 ans)
      1 - Célibataire
      2 - Marié(e) ou remarié(e)
      3 - Veuf(ve)
      4 - Divorcé(e)
    """
    assert individus.matri.isin(range(5)).all()

    individus['statut_marital'] = 2  # célibataire par défaut
    individus.loc[individus.matri == 2, 'statut_marital'] = 1  # marié(e)
    individus.loc[individus.matri == 3, 'statut_marital'] = 4  # veuf(ve)
    individus.loc[individus.matri == 4, 'statut_marital'] = 3  # divorcé(e)

    assert individus.statut_marital.isin(range(1, 7)).all()


def create_taux_csg_remplacement(individus, period, tax_benefit_system, sigma = (28.1) ** 2):
    assert 'revkire' in individus
    assert 'nbp' in individus
    if period.start.year < 2015:  # Should be an assert
        period = periods.period(2015)

    rfr = individus.revkire
    nbptr = individus.nbp / 100

    def compute_taux_csg_remplacement(rfr, nbptr):
        parameters = tax_benefit_system.get_parameters_at_instant(period.start)
        seuils = parameters.prelevements_sociaux.contributions_sociales.csg.remplacement.pensions_de_retraite_et_d_invalidite
        seuil_exoneration = seuils.seuil_de_rfr_1 + (nbptr - 1) * seuils.demi_part_suppl
        seuil_reduction = seuils.seuil_de_rfr_2 + (nbptr - 1) * seuils.demi_part_suppl
        taux_csg_remplacement = 0.0 * rfr
        taux_csg_remplacement = np.where(
            rfr <= seuil_exoneration,
            1,
            np.where(
                rfr <= seuil_reduction,
                2,
                3,
                )
            )

        return taux_csg_remplacement

    individus['rfr_special_csg_n'] = rfr
    individus['taux_csg_remplacement'] = compute_taux_csg_remplacement(rfr, nbptr)
    if sigma is not None:
        np.random.seed(42)
        rfr_n_1 = rfr + (individus.retraite_imposable > 0) * np.random.normal(
            scale = sigma, size = len(individus['taux_csg_remplacement'])
            )
    individus['rfr_special_csg_n_1'] = rfr_n_1
    individus['taux_csg_remplacement_n_1'] = compute_taux_csg_remplacement(rfr_n_1, nbptr)

    distribution = individus.groupby(['taux_csg_remplacement', 'taux_csg_remplacement_n_1'])['ponderation'].sum() / 1000
    log.debug("Distribution of taux_csg_remplacement (in thousands):\n", distribution)
    assert individus['taux_csg_remplacement_n_1'].isin(range(4)).all()
    assert individus['taux_csg_remplacement'].isin(range(4)).all()


def calibrate_categorie_salarie(individus, year = None, mass_by_categorie_salarie = None):
    assert mass_by_categorie_salarie is not None
    log.debug('Mass of salaries: ', mass_by_categorie_salarie)

    weight_individus = individus['ponderation'].values
    for rebalanced_categorie, target_mass in mass_by_categorie_salarie.items():
        categorie_salarie = individus['categorie_salarie'].values
        eligible = (
            (categorie_salarie == 0) | (categorie_salarie == rebalanced_categorie)
            )
        take = (categorie_salarie == rebalanced_categorie)
        log.info(
            """
initial take population: {}
initial eligible population: {}
target mass: {}""".format(
                (take * eligible * weight_individus).sum() / 1e6,
                (eligible * weight_individus).sum() / 1e6,
                target_mass / 1e6,
                ))
        selected = select_to_match_target(
            target_mass = target_mass,
            eligible = eligible,
            weights = weight_individus,
            take = take,
            seed = 9779972
            )
        log.debug("""
    final selected population: {}
    error: {} %
    """.format(
            (eligible * selected * weight_individus).sum() / 1e6,
            ((eligible * selected * weight_individus).sum() - target_mass) / target_mass * 100,
            ))
        individus.loc[selected, 'categorie_salarie'] = rebalanced_categorie
        log.debug(individus.groupby('categorie_salarie')['ponderation'].sum())
        seuil_salaire_imposable_mensuel = 2 * 3000
        individus.loc[
            (
                (individus.contrat_de_travail == 0)
                & (individus.categorie_salarie == 0)
                & (individus.salaire_imposable > 12 * seuil_salaire_imposable_mensuel)
                ),
            'categorie_salarie'
            ] = 1
        individus.loc[
            (
                (individus.contrat_de_travail == 1)
                & (individus.categorie_salarie == 0)
                & (individus.salaire_imposable  > (12 * seuil_salaire_imposable_mensuel) / (35 * 52) * individus.heures_remunerees_volume)
                ),
            'categorie_salarie'
            ] = 1


def todo_create(individus):
    txtppb = "txtppb" if "txtppb" in individus.columns else "txtppred"
    log.debug("6.3 : variable txtppb")
    individus.loc[individus.txtppb.isnull(), txtppb] = 0
    individus.loc[individus[txtppb] == 9, txtppb] = 0
    assert individus.txtppb.notnull().all()
    log.debug("Valeurs prises par la variable txtppb \n {}".format(
        individus[txtppb].value_counts(dropna = False)))


if __name__ == '__main__':

    import sys
    logging.basicConfig(level = logging.INFO, stream = sys.stdout)
    # logging.basicConfig(level = logging.INFO,  filename = 'step_03.log', filemode = 'w')
    year = 2014

    #    from openfisca_france_data.erfs_fpr.input_data_builder import step_01_preprocessing
    #    step_01_preprocessing.build_merged_dataframes(year = year)
    individus = build_variables_individuelles(year = year)



#! /usr/bin/env python

import argparse
import logging
import time
import sys

from openfisca_france_data.erfs.input_data_builder import build as build_erfs
from openfisca_france_data.erfs_fpr.input_data_builder import build as build_erfs_fpr

log = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--collection',
         help = "name of collection to build or update (erfs or erfs_fpr)", required = True)
    parser.add_argument('-d', '--replace-data', action = 'store_true', default = False,
        help = "erase existing survey data HDF5 file (instead of failing when HDF5 file already exists)")
    parser.add_argument('-s', '--survey', help = 'name of survey (year) to build or update', required = True)
    parser.add_argument('-v', '--verbose', action = 'store_true', default = False, help = "increase output verbosity")
    args = parser.parse_args()
    logging.basicConfig(level = logging.DEBUG if args.verbose else logging.WARNING, stream = sys.stdout)
    year = int(args.survey)
    start = time.time()
    if args.collection == 'erfs_fpr':
        build_erfs_fpr(year = year, check = False)
    elif args.collection == 'erfs':
        build_erfs(year = year, check = False)

    log.info("Build lasted {}".format(time.time() - start))


if __name__ == '__main__':
    sys.exit(main())



import logging
import os

import numpy
from pandas import Series

import openfisca_france  # type: ignore
from openfisca_survey_manager.survey_collections import SurveyCollection  # type: ignore
from openfisca_survey_manager.surveys import Survey  # type: ignore

from openfisca_france_data import openfisca_france_tax_benefit_system


log = logging.getLogger(__name__)


def assert_dtype(series, dtype_string):
    assert isinstance(series, Series), "First argument is not of Series type"
    try:
        assert series.dtype == numpy.dtype(dtype_string), "Series {} dtype is {} instead of {}".format(
            series.name, series.dtype, dtype_string)
    except AssertionError:
        assert numpy.issubdtype(series.dtype, numpy.dtype(dtype_string)), "Series {} dtype is {} instead of {}".format(
            series.name, series.dtype, dtype_string)


def assert_variable_in_range(name, wrange, table):
    """Asserts if transformed variables are in correct range.

    wrange is a list like [minimum, maximum]
    """
    temp = (table[table[name].notnull()])
    range_1 = wrange[0]
    range_2 = wrange[1]
    for v in temp[name]:
        assert v in range(range_1, range_2), 'some non-null values for %s not in wanted %s: %s' % (
            name, str(wrange), str(v))


def control(dataframe, verbose = False, verbose_columns = None, debug = False, verbose_length = 5, ignore = None):
    """Helps debugging the data crunchin' files.

    Parameters
    ---------
    verbose: Default False
    Indicates whether to print the dataframe itself or just perform reguler checks.

    verbose_columns: List
    The columns of the dataframe to print

    verbose_length: Int
    the number of rows to print
    """
    std_list = ['idfoy', 'quifoy', 'idmen', 'quimen', 'idfam', 'quifam']
    for var in std_list:
        try:
            assert var in dataframe.columns
        except Exception:
            raise Exception('the dataframe does not contain the required column %s' % var)

    log.info('longueur de la data frame = {}'.format(len(dataframe.index)))
    if debug:
        log.info('nb de doublons: {}'.format(len(dataframe[dataframe.duplicated()])))
        log.info('nb de doublons idfoy/quifoy: {}'.format(len(dataframe[dataframe.duplicated(
            subset = ['idfoy', 'quifoy'])])))
        log.info('nb de doublons idmen/quimen: {}'.format(len(dataframe[dataframe.duplicated(
            subset = ['idmen', 'quimen'])])))
        log.info('nb de doublons idfam/quifam: {}'.format(len(dataframe[dataframe.duplicated(
            subset = ['idfam', 'quifam'])])))

    if not(debug):
        assert not(dataframe.duplicated().any()), 'présence de lignes en double dans la dataframe'
        assert ~(dataframe.duplicated(subset = ['idfoy', 'quifoy'])).all(), 'duplicate of tuple idfoy/quifoy'
        assert ~(dataframe.duplicated(subset = ['idmen', 'quimen'])).all(), 'duplicate of tuple idmen/quimen'
        assert ~(dataframe.duplicated(subset = ['idfam', 'quifam'])).all(), 'duplicate of tuple idfam/quifam'

    empty_columns = []
    for col in dataframe:
        if dataframe[col].isnull().all():
            empty_columns.append(col)

    if empty_columns != []:
        log.info('liste des colonnes entièrement vides: \n {}'.format(empty_columns))

    if verbose is True:
        log.info('------ informations détaillées -------')
        print_id(dataframe)

        if verbose_columns is None:
            if dataframe.duplicated().any():
                log.info(dataframe[dataframe.duplicated()].head(verbose_length).to_string())

        else:
            if dataframe.duplicated(verbose_columns).any():
                print('nb lignes lignes dupliquées _____', len(dataframe[dataframe.duplicated(verbose_columns)]))
                print(dataframe.loc[:, verbose_columns].describe())
            for col in verbose_columns:
                print('nombre de NaN dans %s : ' % (col), dataframe[col].isnull().sum())
            print('colonnes contrôlées ------>', verbose_columns)
    print('vérifications terminées')


def count_NA(name, table):
    """Counts the number of Na's in a specified axis."""
    print("count of NA's for %s is %s" % (name, str(sum(table[name].isnull()))))


def id_formatter(dataframe, entity_id):
    dataframe[entity_id + "_original"] = dataframe[entity_id].copy()
    id_unique = dataframe[entity_id].unique()
    new_id_by_old_id = dict(zip(
        id_unique, range(len(id_unique)))
        )
    dataframe[entity_id] = dataframe[entity_id].map(new_id_by_old_id)
    return dataframe


def print_id(df):
    try:
        log.info("Individus with distinc noindiv: {} / {}".format(len(df.noindiv), len(df)))
    except Exception:
        log.info("No noindiv")

    try:
        # Ici, il doit y avoir autant de vous que d'idfoy
        log.info("Individus dans les Foyers: {} \n".format(len(df.idfoy)))
        log.info(df["quifoy"].value_counts(dropna=False))
        if df["idfoy"].isnull().any():
            log.info("NaN in idfoy : {}".format(df["idfoy"].isnull().sum()))
        if df["quifoy"].isnull().any():
            log.info("NaN in quifoy : {}".format(df["quifoy"].isnull().sum()))
    except Exception:
        log.info("No idfoy or quifoy")

    try:
        # Ici, il doit y avoir autant de quimen = 0 que d'idmen
        log.info(u"Individus dans les Ménages {} \n".format(len(df.idmen)))
        log.info("\n" + str(df["quimen"].value_counts(dropna=False)))
        if df["idmen"].isnull().any():
            log.info("NaN in idmen : {} ".format(df["idmen"].isnull().sum()))
        if df["quimen"].isnull().any():
            log.info("NaN in quimen : {} ".format(df["quimen"].isnull().sum()))
    except Exception:
        print("No idmen or quimen")

    try:
        # Ici, il doit y avoir autant de quifam = 0 que d'idfam
        log.info("Individus dans les Familles {}\n".format(len(df.idfam)))
        log.info(df["quifam"].value_counts(dropna=False))
        if df["idfam"].isnull().any():
            log.info("NaN in idfam : {} ".format(df["idfam"].isnull().sum()))
        if df["quifam"].isnull().any():
            log.info("NaN in quifam : {} ".format(df["quifam"].isnull().sum()))
    except Exception:
        log.info("No idfam or quifam")
    compute_masses(df)


def compute_masses(dataframe):
    variables = [
        'chomage_imposable',
        'pensions_alimentaires_percues',
        'rag',
        'retraite_imposable',
        'ric',
        'rnc',
        'salaire_imposable',
        'f4ba',
        ]
    for variable in variables:
        if set([variable, 'wprm']).issubset(set(dataframe.columns)):
            log.info("Mass of {}: {}".format(variable, (dataframe[variable] * dataframe['wprm']).sum() / 1e9))
        else:
            log.info("Impossible to compute mass of {}".format(variable))


def check_entity_structure(dataframe, entity):
    log.info("Checking entity {}".format(entity))
    role = 'qui' + entity
    entity_id = 'id' + entity
    error_messages = list()

    if dataframe[role].isnull().any():
        error_messages.append("there are NaN in qui{}".format(entity))
    max_entity_role_value = dataframe[role].max().astype("int")
    id_count = len(dataframe['id' + entity].unique())
    head_count = (dataframe['qui' + entity] == 0).sum()
    if id_count != head_count:
        error_messages.append("Wrong number of heads for {}: {} different ids for {} heads".format(
            entity, id_count, head_count))

    entity_ids_by_role = dict()
    for role_value in range(max_entity_role_value + 1, 1, -1):
        print(entity, role_value)
        log.info("Dealing with role {} of entity {}".format(role_value, entity))
        entity_ids_by_role[role_value] = set(dataframe.loc[dataframe[role] == role_value, entity_id].unique())

        if role_value < max_entity_role_value:
            difference = entity_ids_by_role[role_value + 1].difference(entity_ids_by_role[role_value])
            if not entity_ids_by_role[role_value + 1].issubset(entity_ids_by_role[role_value]):
                error_messages.append(
                    "Problem with entity {} at role = {} for id {}".format(
                        entity, role_value, difference
                        )
                    )
                erroneous_ids = difference
                return False, error_messages, erroneous_ids
            else:
                continue
    return True, None, None


def check_structure(dataframe):
    duplicates = dataframe.noindiv.duplicated().sum()
    messages = list()
    erroneous_ids_by_entity = dict()
    if duplicates != 0:
        messages.append("There are {} duplicated individuals".format(duplicates))
    for entity in ['fam', 'foy', 'men']:
        print(entity)
        checked, error_messages, erroneous_ids = check_entity_structure(dataframe, entity)
        print(checked, error_messages, erroneous_ids)
        if not checked:
            messages.append('Structure error for {}'.format(entity))
            messages.append(error_messages)
            erroneous_ids_by_entity[entity] = erroneous_ids
    if not messages:
        return True, None
    else:
        log.info('\n'.join('{}'.format(item) for item in messages))
        return False, erroneous_ids_by_entity


def build_cerfa_fields_by_column_name(year, sections_cerfa):
    tax_benefit_system = openfisca_france_tax_benefit_system
    cerfa_fields_by_column_name = dict()
    for name, column in tax_benefit_system.variables.items():
        for section_cerfa in sections_cerfa:
            if name.startswith(f"f{section_cerfa}"):
                end = column.end or None

                if end is None or end.year >= year:
                    if column.entity.key == 'individu':
                        cerfa_field = ['f' + x.lower() for x in list(column.cerfa_field.values())]
                    elif column.entity.key == 'foyer_fiscal':
                        cerfa_field = ['f' + column.cerfa_field.lower()]
                    cerfa_fields_by_column_name[name] = cerfa_field
    return cerfa_fields_by_column_name


def build_cerfa_fields_by_variable(year):
    tax_benefit_system = openfisca_france_tax_benefit_system
    cerfa_fields_by_variable = dict()
    for name, variable in sorted(tax_benefit_system.variables.items()):
        if variable.cerfa_field is None:
            continue
        if name.startswith('case') or name.startswith('nb'):
            continue
        end = variable.end or None
        if end is None or end.year >= year:
            if variable.entity.key == 'individu':
                cerfa_field = ['f' + x.lower() for x in list(variable.cerfa_field.values())]
            elif variable.entity.key == 'foyer_fiscal':
                cerfa_field = ['f' + variable.cerfa_field.lower()]
            cerfa_fields_by_variable[name] = cerfa_field
    return cerfa_fields_by_variable


def rectify_dtype(dataframe, verbose = True):
    series_to_rectify = []
    rectified_series = []
    for serie_name, serie in dataframe.items():
        if serie.dtype.char == 'O':  # test for object
            series_to_rectify.append(serie_name)
            if verbose:
                print("""
Variable name: {}
NaN are present : {}
{}""".format(serie_name, serie.isnull().sum(), serie.value_counts()))
            # bool
            if serie.dropna().isin([True, False]).all():
                if serie.isnull().any():
                    serie = serie.fillna(False).copy()
                dataframe[serie_name] = serie.astype('bool', copy = True)
                rectified_series.append(serie_name)
            # Nombre 01-99
            elif serie.dropna().str.match("\d\d$").all():
                if serie.isnull().any():
                    serie = serie.fillna(0)
                dataframe[serie_name] = serie.astype('int', copy = True)
                rectified_series.append(serie_name)
            # year
            elif serie_name in ['birthvous', 'birthconj', 'caseH']:
                if serie.isnull().any():
                    serie = serie.fillna(9999)
                dataframe[serie_name] = serie.astype('int', copy = True)
                rectified_series.append(serie_name)
            # datetime
            elif serie_name[0:4] == "date":
                from pandas import to_datetime
                dataframe[serie_name] = to_datetime(
                    serie.str[:2] + "-" + serie.str[2:4] + "-" + serie.str[4:8],
                    coerce = True)
                rectified_series.append(serie_name)

            if serie_name in rectified_series:
                if verbose:
                    print("""Converted to {}
{}""".format(dataframe[serie_name].dtype, dataframe[serie_name].value_counts()))

    if verbose:
        print(set(series_to_rectify).difference(rectified_series))


def normalizes_roles_in_entity(dataframe, entity_id_name, entity_role_name):
    last_roles = (dataframe[entity_role_name] >= 2).copy()
    dataframe.loc[last_roles, entity_role_name] = 2
    dataframe[entity_role_name] = dataframe[entity_role_name].astype('int')


def set_variables_default_value(dataframe, year):
    tax_benefit_system = openfisca_france_tax_benefit_system

    for column_name, column in tax_benefit_system.variables.items():
        if column_name in dataframe.columns:
            dataframe[column_name].fillna(column.default_value, inplace = True)
            dataframe[column_name] = dataframe[column_name].astype(column.dtype)


def store_input_data_frame(data_frame = None, collection = None, survey = None, table = None):
    assert data_frame is not None
    assert collection is not None
    assert survey is not None
    try:
        openfisca_survey_collection = SurveyCollection.load(collection = collection)
    except Exception as e:
        openfisca_survey_collection = SurveyCollection(name = collection)

    log.debug("In collection {} the following survey are present: {}".format(collection, openfisca_survey_collection.surveys))
    output_data_directory = openfisca_survey_collection.config.get('data', 'output_directory')
    if table is None:
        table = "input"
    #
    survey_name = survey
    hdf5_file_path = os.path.join(os.path.dirname(output_data_directory), "{}.h5".format(survey_name))
    available_survey_names = [survey_.name for survey_ in openfisca_survey_collection.surveys]
    if survey_name in available_survey_names:
        survey = openfisca_survey_collection.get_survey(survey_name)
    else:
        survey = Survey(name = survey_name, hdf5_file_path = hdf5_file_path)
    survey.insert_table(name = table, data_frame = data_frame)
    openfisca_survey_collection.surveys.append(survey)
    collections_directory = openfisca_survey_collection.config.get('collections', 'collections_directory')
    json_file_path = os.path.join(collections_directory, '{}.json'.format(collection))
    log.debug("In collection {} the following surveyx are present: {}".format(collection, openfisca_survey_collection.surveys))
    openfisca_survey_collection.dump(json_file_path = json_file_path)
